# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Detector functionality of SpeciesNet."""

__all__ = [
    "SpeciesNetDetector",
]

import time
from typing import Any, Optional

from absl import logging
from humanfriendly import format_timespan
import numpy as np
import PIL.Image
import torch
import torch.backends
import torch.backends.mps
from yolov5.utils.augmentations import letterbox as yolov5_letterbox
from yolov5.utils.general import non_max_suppression as yolov5_non_max_suppression
from yolov5.utils.general import scale_boxes as yolov5_scale_boxes
from yolov5.utils.general import xyxy2xywhn as yolov5_xyxy2xywhn

from speciesnet.constants import Detection
from speciesnet.constants import Failure
from speciesnet.utils import ModelInfo
from speciesnet.utils import PreprocessedImage


class SpeciesNetDetector:
    """Detector component of SpeciesNet."""

    IMG_SIZE = 1280
    STRIDE = 64
    DETECTION_THRESHOLD = 0.01

    def __init__(self, model_name: str) -> None:
        """Loads the detector resources.

        Code adapted from: https://github.com/agentmorris/MegaDetector
        which was released under the MIT License:
        https://github.com/agentmorris/MegaDetector/blob/main/LICENSE

        Args:
            model_name:
                String value identifying the model to be loaded. It can be a Kaggle
                identifier (starting with `kaggle:`), a HuggingFace identifier (starting
                with `hf:`) or a local folder to load the model from.
        """

        start_time = time.time()

        self.model_info = ModelInfo(model_name)

        # Select the best device available.
        if torch.cuda.is_available():
            self.device = "cuda"
        elif torch.backends.mps.is_available():
            self.device = "mps"
        else:
            self.device = "cpu"

        # Load the model.
        if self.device != "mps":
            checkpoint = torch.load(
                self.model_info.detector, map_location=self.device, weights_only=False
            )
            self.model = checkpoint["model"].float()
        else:
            checkpoint = torch.load(self.model_info.detector, weights_only=False)
            self.model = checkpoint["model"].float().to(self.device)

        # Set the model in inference mode.
        self.model.eval()
        for param in self.model.parameters():
            param.requires_grad = False

        # Fix compatibility issues to be able to load older YOLOv5 models with newer
        # versions of PyTorch.
        for m in self.model.modules():
            if isinstance(m, torch.nn.Upsample) and not hasattr(
                m, "recompute_scale_factor"
            ):
                m.recompute_scale_factor = None

        end_time = time.time()
        logging.info(
            "Loaded SpeciesNetDetector in %s on %s.",
            format_timespan(end_time - start_time),
            self.device.upper(),
        )

    def preprocess(self, img: Optional[PIL.Image.Image]) -> Optional[PreprocessedImage]:
        """Preprocesses an image according to this detector's needs.

        Args:
            img:
                PIL image to preprocess. If `None`, no preprocessing is performed.

        Returns:
            A preprocessed image, or `None` if no PIL image was provided initially.
        """

        if img is None:
            return None

        img_arr = yolov5_letterbox(
            np.asarray(img),
            new_shape=SpeciesNetDetector.IMG_SIZE,
            stride=SpeciesNetDetector.STRIDE,
            auto=True,
        )[0]
        return PreprocessedImage(img_arr, img.width, img.height)

    def _convert_yolo_xywhn_to_md_xywhn(self, yolo_xywhn: list[float]) -> list[float]:
        """Converts bbox XYWHN coordinates from YOLO's to MegaDetector's format.

        Args:
            yolo_xywhn:
                List of bbox coordinates in YOLO format, i.e.
                [x_center, y_center, width, height].

        Returns:
            List of bbox coordinates in MegaDetector format, i.e.
            [x_min, y_min, width, height].
        """

        x_center, y_center, width, height = yolo_xywhn
        x_min = x_center - width / 2.0
        y_min = y_center - height / 2.0
        return [x_min, y_min, width, height]

    def predict(
        self, filepath: str, img: Optional[PreprocessedImage]
    ) -> dict[str, Any]:
        """Runs inference on a given preprocessed image.

        Code adapted from: https://github.com/agentmorris/MegaDetector
        which was released under the MIT License:
        https://github.com/agentmorris/MegaDetector/blob/main/LICENSE

        Args:
            filepath:
                Location of image to run inference on. Used for reporting purposes only,
                and not for loading the image.
            img:
                Preprocessed image to run inference on. If `None`, a failure message is
                reported back.

        Returns:
            A dict containing either the detections above a fixed confidence threshold
            for the given image (in decreasing order of confidence scores), or a failure
            message if no preprocessed image was provided.
        """

        if img is None:
            return {
                "filepath": filepath,
                "failures": [Failure.DETECTOR.name],
            }

        # Prepare model input.
        img_tensor = torch.from_numpy(img.arr / 255)
        img_tensor = img_tensor.permute([2, 0, 1])  # HWC to CHW.
        batch_tensor = torch.unsqueeze(img_tensor, 0).float()  # CHW to NCHW.
        batch_tensor = batch_tensor.to(self.device)

        # Run inference.
        results = self.model(batch_tensor, augment=False)[0]
        if self.device == "mps":
            results = results.cpu()
        results = yolov5_non_max_suppression(
            prediction=results,
            conf_thres=SpeciesNetDetector.DETECTION_THRESHOLD,
        )
        results = results[0]  # Drop batch dimension.

        # Process detections.
        detections = []
        results[:, :4] = yolov5_scale_boxes(
            batch_tensor.shape[2:],
            results[:, :4],
            (img.orig_height, img.orig_width),
        ).round()
        for result in results:  # (x_min, y_min, x_max, y_max, conf, category)
            xyxy = result[:4]
            xywhn = yolov5_xyxy2xywhn(xyxy, w=img.orig_width, h=img.orig_height)
            bbox = self._convert_yolo_xywhn_to_md_xywhn(xywhn.tolist())

            conf = result[4].item()

            category = str(int(result[5].item()) + 1)
            label = Detection.from_category(category)
            if label is None:
                logging.error("Invalid detection class: %s", category)
                continue

            detections.append(
                {
                    "category": category,
                    "label": label.value,
                    "conf": conf,
                    "bbox": bbox,
                }
            )

        # Sort detections by confidence score.
        detections = sorted(detections, key=lambda det: det["conf"], reverse=True)

        return {
            "filepath": filepath,
            "detections": detections,
        }



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualization utilities."""

__all__ = [
    "draw_bboxes",
]

from matplotlib import font_manager
import PIL
import PIL.Image
import PIL.ImageColor
import PIL.ImageDraw
import PIL.ImageFile
import PIL.ImageFont

from speciesnet.constants import Detection
from speciesnet.utils import BBox

# Color map for different detection types.
DETECTIONS_COLOR_MAP = {
    Detection.ANIMAL.value: "red",
    Detection.HUMAN.value: "blue",
    Detection.VEHICLE.value: "magenta",
}


def draw_bboxes(img: PIL.Image.Image, detections: list[dict]) -> PIL.Image.Image:
    """Draws bounding boxes on a given image.

    The detections are assumed to be in decreasing order of their confidence score. They
    are drawn from the least confident to the most confident, to make sure that the
    latter is always on top. The opacity of each bounding box is proportional to its
    confidence score, so the least confident boxes may sometimes be barely visible,
    indicating that low confidence.

    Args:
        img:
            PIL image to draw bounding boxes on.
        detections:
            List of detections sorted in decreasing order of their confidence score.

    Returns:
        A new PIL image generated by overlaying detection bounding boxes on top of the
        original PIL image.
    """

    font_prop = font_manager.FontProperties(family="sans serif", weight="bold")
    font = PIL.ImageFont.truetype(font_manager.findfont(font_prop), size=12)
    border_size = 3

    img = img.convert("RGBA")
    overlay = PIL.Image.new("RGBA", img.size, (255, 255, 255, 0))
    draw = PIL.ImageDraw.Draw(overlay)

    # Draw bounding boxes in reversed order, from the least confident to the most
    # confident, to make sure that the latter is always on top.
    for detection in reversed(detections):
        label = detection["label"]
        score = detection["conf"]
        bbox = BBox(*detection["bbox"])

        x0 = bbox.xmin * img.width
        y0 = bbox.ymin * img.height
        x1 = (bbox.xmin + bbox.width) * img.width
        y1 = (bbox.ymin + bbox.height) * img.height
        rgb = PIL.ImageColor.getrgb(DETECTIONS_COLOR_MAP[label])
        alpha = int(score * 255)
        color = *rgb[:3], alpha

        text = f"{label}: {score:.2f}"
        text_rel_xy = font.getbbox(text, anchor="lb")
        text_bg_xy = (
            text_rel_xy[0] + x0,
            text_rel_xy[1] + y0 - 2 * border_size,
            text_rel_xy[2] + x0 + 2 * border_size,
            text_rel_xy[3] + y0,
        )
        text_color = (255, 255, 255, alpha)

        draw.rectangle((x0, y0, x1, y1), outline=color, width=border_size)
        draw.rectangle(text_bg_xy, fill=color, width=border_size)
        draw.text(
            (x0 + border_size, y0 - border_size),
            text,
            fill=text_color,
            font=font,
            anchor="lb",
        )

    return PIL.Image.alpha_composite(img, overlay)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic for combining predictions for the SpeciesNet ensemble."""

from typing import Callable, Optional

from speciesnet.constants import Classification
from speciesnet.constants import Detection

PredictionLabelType = str
PredictionScoreType = float
PredictionSourceType = str
PredictionType = tuple[PredictionLabelType, PredictionScoreType, PredictionSourceType]


def combine_predictions_for_single_item(
    *,
    classifications: dict[str, list],
    detections: list[dict],
    country: Optional[str],
    admin1_region: Optional[str],
    taxonomy_map: dict,
    geofence_map: dict,
    enable_geofence: bool,
    geofence_fn: Callable,
    roll_up_fn: Callable,
) -> PredictionType:
    """Ensembles classifications and detections for a single image.

    This operation leverages multiple heuristics to make the most of the classifier and
    the detector predictions through a complex set of decisions. It introduces various
    thresholds to identify humans, vehicles, blanks, animals at species level, animals
    at higher taxonomy levels and even unknowns.

    Args:
        classifications:
            Dict of classification results. "classes" and "scores" are expected to be
            provided among the dict keys.
        detections:
            List of detection results, sorted in decreasing order of their confidence
            score. Each detection is expected to be a dict providing "label" and "conf"
            among its keys.
        country:
            Country (in ISO 3166-1 alpha-3 format) associated with predictions.
            Optional.
        admin1_region:
            First-level administrative division (in ISO 3166-2 format) associated with
            predictions. Optional.
        taxonomy_map:
            Dictionary mapping taxa to labels.
        geofence_map:
            Dictionary mapping full class strings to geofence rules.
        enable_geofence:
            Whether geofencing is enabled.
        geofence_fn:
            Callable to geofence animal classifications.
        roll_up_fn:
            Callable to roll up labels to the first matching level.

    Returns:
        A tuple of <label, score, prediction_source> describing the ensemble result.
    """

    top_classification_class = classifications["classes"][0]
    top_classification_score = classifications["scores"][0]
    top_detection_class = detections[0]["label"] if detections else Detection.ANIMAL
    top_detection_score = detections[0]["conf"] if detections else 0.0

    if top_detection_class == Detection.HUMAN:
        # Threshold #1a: high-confidence HUMAN detections.
        if top_detection_score > 0.7:
            return Classification.HUMAN, top_detection_score, "detector"

        # Threshold #1b: mid-confidence HUMAN detections + high-confidence
        # HUMAN/VEHICLE classifications.
        if (
            top_detection_score > 0.2
            and top_classification_class
            in {Classification.HUMAN, Classification.VEHICLE}
            and top_classification_score > 0.5
        ):
            return Classification.HUMAN, top_classification_score, "classifier"

    if top_detection_class == Detection.VEHICLE:
        # Threshold #2a: mid-confidence VEHICLE detections + high-confidence HUMAN
        # classifications.
        if (
            top_detection_score > 0.2
            and top_classification_class == Classification.HUMAN
            and top_classification_score > 0.5
        ):
            return Classification.HUMAN, top_classification_score, "classifier"

        # Threshold #2b: high-confidence VEHICLE detections.
        if top_detection_score > 0.7:
            return Classification.VEHICLE, top_detection_score, "detector"

        # Threshold #2c: mid-confidence VEHICLE detections + high-confidence VEHICLE
        # classifications.
        if (
            top_detection_score > 0.2
            and top_classification_class == Classification.VEHICLE
            and top_classification_score > 0.4
        ):
            return Classification.VEHICLE, top_classification_score, "classifier"

    # Threshold #3a: high-confidence BLANK "detections" + high-confidence BLANK
    # classifications.
    if (
        top_detection_score < 0.2
        and top_classification_class == Classification.BLANK
        and top_classification_score > 0.5
    ):
        return Classification.BLANK, top_classification_score, "classifier"

    # Threshold #3b: extra-high-confidence BLANK classifications.
    if (
        top_classification_class == Classification.BLANK
        and top_classification_score > 0.99
    ):
        return Classification.BLANK, top_classification_score, "classifier"

    if top_classification_class not in {
        Classification.BLANK,
        Classification.HUMAN,
        Classification.VEHICLE,
    }:
        # Threshold #4a: extra-high-confidence ANIMAL classifications.
        if top_classification_score > 0.8:
            return geofence_fn(
                labels=classifications["classes"],
                scores=classifications["scores"],
                country=country,
                admin1_region=admin1_region,
                taxonomy_map=taxonomy_map,
                geofence_map=geofence_map,
                enable_geofence=enable_geofence,
            )

        # Threshold #4b: high-confidence ANIMAL classifications + mid-confidence
        # ANIMAL detections.
        if (
            top_classification_score > 0.65
            and top_detection_class == Detection.ANIMAL
            and top_detection_score > 0.2
        ):
            return geofence_fn(
                labels=classifications["classes"],
                scores=classifications["scores"],
                country=country,
                admin1_region=admin1_region,
                taxonomy_map=taxonomy_map,
                geofence_map=geofence_map,
                enable_geofence=enable_geofence,
            )

    # Threshold #5a: high-confidence ANIMAL rollups.
    rollup = roll_up_fn(
        labels=classifications["classes"],
        scores=classifications["scores"],
        country=country,
        admin1_region=admin1_region,
        target_taxonomy_levels=["genus", "family", "order", "class", "kingdom"],
        non_blank_threshold=0.65,
        taxonomy_map=taxonomy_map,
        geofence_map=geofence_map,
        enable_geofence=enable_geofence,
    )
    if rollup:
        return rollup

    # Threshold #5b: mid-confidence ANIMAL detections.
    if top_detection_class == Detection.ANIMAL and top_detection_score > 0.5:
        return Classification.ANIMAL, top_detection_score, "detector"

    return Classification.UNKNOWN, top_classification_score, "classifier"



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to check GPU availability for PyTorch in the current Python environment."""


def torch_test():
    """Print diagnostic information about Torch/CUDA status, including Torch/CUDA
    versions and all available CUDA device names.
    """

    try:
        import torch
    except Exception as e:
        print(
            "PyTorch unavailable, not running PyTorch tests. "
            "PyTorch import error was:\n{}".format(str(e))
        )
        return

    print("Torch version: {}".format(str(torch.__version__)))
    print("CUDA available (according to PyTorch): {}".format(torch.cuda.is_available()))
    if torch.cuda.is_available():
        print("CUDA version (according to PyTorch): {}".format(torch.version.cuda))  # type: ignore
        print(
            "CuDNN version (according to PyTorch): {}".format(
                torch.backends.cudnn.version()
            )
        )

    device_ids = list(range(torch.cuda.device_count()))

    if len(device_ids) > 0:
        cuda_str = "Found {} CUDA devices:".format(len(device_ids))
        print(cuda_str)

        for device_id in device_ids:
            device_name = "unknown"
            try:
                device_name = torch.cuda.get_device_name(device=device_id)
            except Exception:
                pass
            print("{}: {}".format(device_id, device_name))
    else:
        print("No GPUs reported by PyTorch")

    try:
        if torch.backends.mps.is_built and torch.backends.mps.is_available():
            print("PyTorch reports that Metal Performance Shaders are available")
    except Exception:
        pass
    return len(device_ids)


if __name__ == "__main__":

    print("*** Running Torch tests ***\n")
    torch_test()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utilities."""

__all__ = [
    "BBox",
    "only_one_true",
    "file_exists",
    "load_rgb_image",
    "prepare_instances_dict",
]

from dataclasses import dataclass
from io import BytesIO
import json
from pathlib import Path
import tempfile
from typing import Optional, Union

from absl import logging
from cloudpathlib import CloudPath
from huggingface_hub import snapshot_download
import kagglehub
import numpy as np
import PIL.Image
import PIL.ImageFile
import PIL.ImageOps
import requests

StrPath = Union[str, Path]

PIL.ImageFile.LOAD_TRUNCATED_IMAGES = True

# Supported extensions for image discovery.
IMG_EXTENSIONS = {
    "jpg",
    "JPG",
    "jpeg",
    "JPEG",
    "png",
    "PNG",
    "tif",
    "TIF",
    "tiff",
    "TIFF",
    "webp",
    "WEBP",
}

# Custom agent for image requests over HTTP(S).
CUSTOM_HTTP_AGENT = {"User-Agent": "SpeciesNetBot/0.0 (github.com/google/cameratrapai)"}


@dataclass(frozen=True)
class ModelInfo:
    """Dataclass describing SpeciesNet model and its underlying resources to load."""

    version: str  # Model version.
    type_: str  # Model type.
    classifier: Path  # Path to classifier model.
    classifier_labels: Path  # Path to labels file used by clasifier.
    detector: Path  # Path to detector model.
    taxonomy: Path  # Path to taxonomy file used by ensemble.
    geofence: Path  # Path to geofence file used by ensemble.

    def __init__(self, model_name: str) -> None:
        """Creates dataclass to describe a given model.

        Args:
            model_name:
                String value identifying the model to be described by this dataclass.
                It can be a Kaggle identifier (starting with `kaggle:`), a HuggingFace
                identifier (starting with `hf:`) or a local folder to load the model
                from. If the model name is a remote identifier (Kaggle or HuggingFace),
                the model files are automatically downloaded on the first call.
        """

        # Download model files (if necessary) and set the base local directory.
        kaggle_prefix = "kaggle:"
        hf_prefix = "hf:"
        if model_name.startswith(kaggle_prefix):
            base_dir = kagglehub.model_download(model_name[len(kaggle_prefix) :])
        elif model_name.startswith(hf_prefix):
            base_dir = snapshot_download(model_name[len(hf_prefix) :])
        else:
            base_dir = model_name
        base_dir = Path(base_dir)

        # Load model info.
        with open(base_dir / "info.json", mode="r", encoding="utf-8") as fp:
            info = json.load(fp)

        # Download detector weights if not provided with the other model files.
        filepath_or_url = info["detector"]
        if filepath_or_url.startswith("http://") or filepath_or_url.startswith(
            "https://"
        ):
            filename = self._url_to_filename(filepath_or_url)
            info["detector"] = filename
            filepath = base_dir / filename
            if not filepath.exists():
                response = requests.get(filepath_or_url, stream=True, timeout=600)
                response.raise_for_status()
                with open(filepath, mode="wb") as fp:
                    for chunk in response.iter_content(chunk_size=8192):
                        fp.write(chunk)

        # Set dataclass fields using a workaround to bypass read-only constraints.
        object.__setattr__(self, "version", info["version"])
        object.__setattr__(self, "type_", info["type"])
        object.__setattr__(self, "classifier", base_dir / info["classifier"])
        object.__setattr__(
            self, "classifier_labels", base_dir / info["classifier_labels"]
        )
        object.__setattr__(self, "detector", base_dir / info["detector"])
        object.__setattr__(self, "taxonomy", base_dir / info["taxonomy"])
        object.__setattr__(self, "geofence", base_dir / info["geofence"])

    def _url_to_filename(self, url: str) -> str:
        """Sanitizes a URL to get a valid filename.

        Args:
            url: String value for the URL to sanitize.

        Returns:
            String value representing a valid filename obtained from the sanitized URL.
        """

        if "?" in url:
            filename = url.split("?")[0]
        else:
            filename = url
        filename = filename.replace(":", "_").replace("/", "_")
        return filename


@dataclass(frozen=True)
class PreprocessedImage:
    """Dataclass describing a preprocessed image."""

    arr: np.ndarray  # Multidimensional array of image pixels.
    orig_width: int  # Original image width.
    orig_height: int  # Original image height.


@dataclass(frozen=True)
class BBox:
    """Dataclass describing a bounding box."""

    xmin: float
    ymin: float
    width: float
    height: float


def only_one_true(*args) -> bool:
    """Checks that only one of the given arguments is `True`."""

    already_found = False
    for arg in args:
        if arg:
            if already_found:
                return False
            else:
                already_found = True
    return already_found


def file_exists(filepath_or_url: str) -> bool:
    """Checks whether a given file exists and is accessible.

    Args:
        filepath_or_url:
            String representing either a local file, an `http(s)://` URL or a cloud
            location (identified by one of these prefixes: `az://`, `gs://`, `s3://`).

    Returns:
        `True` if file exists and is accessible, or `False` otherwise.
    """

    try:
        if filepath_or_url.startswith("http://") or filepath_or_url.startswith(
            "https://"
        ):
            return requests.get(
                filepath_or_url, headers=CUSTOM_HTTP_AGENT, timeout=60
            ).ok
        elif (
            filepath_or_url.startswith("az://")
            or filepath_or_url.startswith("gs://")
            or filepath_or_url.startswith("s3://")
        ):
            return CloudPath(filepath_or_url).exists()  # type: ignore
        else:
            return Path(filepath_or_url).exists()
    except Exception as e:  # pylint: disable=broad-exception-caught
        logging.error(
            "`%s` while loading `%s` ==> %s", type(e).__name__, filepath_or_url, e
        )
        return False


def load_rgb_image(filepath_or_url: str) -> Optional[PIL.Image.Image]:
    """Loads a file as an RGB PIL image.

    Args:
        filepath_or_url:
            String representing either a local file, an `http(s)://` URL or a cloud
            location (identified by one of these prefixes: `az://`, `gs://`, `s3://`).

    Returns:
        An RGB PIL image if the file was loaded successfully, or `None` otherwise.
    """

    try:
        if filepath_or_url.startswith("http://") or filepath_or_url.startswith(
            "https://"
        ):
            file_contents = requests.get(
                filepath_or_url, headers=CUSTOM_HTTP_AGENT, timeout=60
            ).content
            img = PIL.Image.open(BytesIO(file_contents))
        else:
            if (
                filepath_or_url.startswith("az://")
                or filepath_or_url.startswith("gs://")
                or filepath_or_url.startswith("s3://")
            ):
                path = CloudPath(filepath_or_url)  # type: ignore
            else:
                path = Path(filepath_or_url)
            img = PIL.Image.open(path)

        img.load()
        img = img.convert("RGB")
        img = PIL.ImageOps.exif_transpose(img)
        return img

    except Exception as e:  # pylint: disable=broad-exception-caught
        logging.error(
            "`%s` while loading `%s` ==> %s", type(e).__name__, filepath_or_url, e
        )
        return None


def prepare_instances_dict(  # pylint: disable=too-many-positional-arguments
    instances_dict: Optional[dict] = None,
    instances_json: Optional[StrPath] = None,
    filepaths: Optional[list[StrPath]] = None,
    filepaths_txt: Optional[StrPath] = None,
    folders: Optional[list[StrPath]] = None,
    folders_txt: Optional[StrPath] = None,
    country: Optional[str] = None,
    admin1_region: Optional[str] = None,
) -> dict:
    """Transforms various input formats into an instances dict.

    The instances dict is the most expressive input format of them all since, compared
    to others, it can also express country, latitude and longitude information.

    This function expects that only one input argument is provided.

    Args:
        instances_dict:
            Optional instances dict. If provided, this function is a no-op.
        instances_json:
            Optional path to load the instances dict from.
        filepaths:
            Optional list of filepaths to process.
        filepaths_txt:
            Optional path to load the list of filepaths to process from.
        folders:
            Optional list of folders to process.
        folders_txt:
            Optional path to load the list of folders to process from.
        country:
            Optional country to enforce on all instances.
        admin1_region:
            Optional first-level administrative division to enforce on all instances.

    Returns:
        An instances dict resulted from the input transformation.

    Raises:
        ValueError:
            If more than one input argument was provided.
    """

    def _enforce_location(
        instances_dict: dict, country: Optional[str], admin1_region: Optional[str]
    ) -> dict:
        if not country:
            return instances_dict
        location_dict = {"country": country}
        if admin1_region:
            location_dict["admin1_region"] = admin1_region
        return {
            "instances": [
                instance_dict | location_dict
                for instance_dict in instances_dict["instances"]
            ]
        }

    inputs_str = (
        "["
        "instances_dict, "
        "instances_json, "
        "filepaths, "
        "filepaths_txt, "
        "folders, "
        "folders_txt"
        "]"
    )
    inputs = eval(inputs_str)  # pylint: disable=eval-used
    if not only_one_true(*inputs):
        raise ValueError(
            f"Expected exactly one of {inputs_str} to be provided. "
            f"Received: {inputs}."
        )

    if instances_json is not None:
        with open(instances_json, mode="r", encoding="utf-8") as fp:
            instances_dict = json.load(fp)
    if instances_dict is not None:
        return _enforce_location(instances_dict, country, admin1_region)

    if folders_txt is not None:
        with open(folders_txt, mode="r", encoding="utf-8") as fp:
            folders = [line.strip() for line in fp.readlines()]
    if folders is not None:
        filepaths = []
        for folder in folders:
            base_dir = Path(folder)
            for ext in IMG_EXTENSIONS:
                filepaths.extend(base_dir.glob(f"**/*.{ext}"))
        filepaths = sorted(set(filepaths))

    if filepaths_txt is not None:
        with open(filepaths_txt, mode="r", encoding="utf-8") as fp:
            filepaths = [line.strip() for line in fp.readlines()]
    assert filepaths is not None
    return _enforce_location(
        {
            "instances": [
                {
                    "filepath": (
                        filepath if isinstance(filepath, str) else filepath.as_posix()
                    )
                }
                for filepath in filepaths
            ]
        },
        country,
        admin1_region,
    )


def load_partial_predictions(
    predictions_json: Optional[StrPath],
    instances: list[dict],
) -> tuple[dict[str, dict], list[dict]]:
    """Loads partial predictions and filters unprocessed instances from a given list.

    Args:
        predictions_json:
            Filepath to partial predictions to load. If missing, no previous predictions
            can be reused and all instances are considered unprocessed.
        instances:
            List of instances to check if they still need to be processed. Those found
            in the partial predictions are considered already processed and are filtered
            out.

    Returns:
        A tuple made of: (a) the partial predictions dict, and (b) the list of
        unprocessed instances.

    Raises:
        RuntimeError:
            If the partial predictions contain a filepath not found in the list of
            instances to check. To fix this, make sure that the partial predictions
            originated from the same list of instances to process.
    """

    if not predictions_json:
        return {}, instances
    predictions_json = Path(predictions_json)
    if not predictions_json.exists() or not predictions_json.is_file():
        return {}, instances

    logging.info("Loading partial predictions from `%s`.", predictions_json)

    partial_predictions = {}
    target_filepaths = {instance["filepath"] for instance in instances}
    with open(predictions_json, mode="r", encoding="utf-8") as fp:
        predictions_dict = json.load(fp)
        for prediction in predictions_dict["predictions"]:
            filepath = prediction["filepath"]
            if filepath not in target_filepaths:
                raise RuntimeError(
                    f"Filepath from loaded predictions is missing from the set of "
                    f"instances to process: `{filepath}`. Make sure you're resuming "
                    f"the work using the same set of instances."
                )

            if "failures" in prediction:
                continue

            partial_predictions[prediction["filepath"]] = prediction

    instances_to_process = [
        instance
        for instance in instances
        if instance["filepath"] not in partial_predictions
    ]

    return partial_predictions, instances_to_process


def save_predictions(predictions_dict: dict, output_json: StrPath) -> None:
    """Saves a predictions dict to an output file.

    Args:
        predictions_dict:
            Predictions dict to save.
        output_json:
            Output filepath where to save the predictions dict in JSON format.
    """

    output_json = Path(output_json)
    with tempfile.NamedTemporaryFile(
        mode="w",
        dir=output_json.parent,
        prefix=f"{output_json.name}.tmp.",
        delete=False,
    ) as fp:
        logging.info("Saving predictions to `%s`.", fp.name)
        output_json_tmp = Path(fp.name)
        json.dump(predictions_dict, fp, ensure_ascii=False, indent=4)
    logging.info("Moving `%s` to `%s`.", output_json_tmp, output_json)
    output_json_tmp.replace(output_json)  # Atomic operation.



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to convert SpeciesNet .json format to MegaDetector/Timelapse .json format."""

import argparse
import json
import sys

blank_prediction_string = "f1856211-cfb7-4a5b-9158-c0f72fd09ee6;;;;;;blank"
no_cv_result_prediction_string = "f2efdae9-efb8-48fb-8a91-eccf79ab4ffb;no cv result;no cv result;no cv result;no cv result;no cv result;no cv result"
rodent_prediction_string = (
    "90d950db-2106-4bd9-a4c1-777604c3eada;mammalia;rodentia;;;;rodent"
)
mammal_prediction_string = "f2d233e3-80e3-433d-9687-e29ecc7a467a;mammalia;;;;;mammal"
animal_prediction_string = "1f689929-883d-4dae-958c-3d57ab5b6c16;;;;;;animal"
human_prediction_string = "990ae9dd-7a59-4344-afcb-1b7b21368000;mammalia;primates;hominidae;homo;sapiens;human"


def invert_dictionary(d):
    """
    Creates a new dictionary that maps d.values() to d.keys().  Does not check
    uniqueness.

    Args:
        d (dict): dictionary to invert

    Returns:
        dict: inverted copy of [d]
    """

    return {v: k for k, v in d.items()}


def sort_list_of_dicts_by_key(L, k, reverse=False):
    """
    Sorts the list of dictionaries [L] by the key [k].

    Args:
        L (list): list of dictionaries to sort
        k (object, typically str): the sort key
        reverse (bool, optional): whether to sort in reverse (descending) order

    Returns:
        dict: sorted copy of [d]
    """
    return sorted(L, key=lambda d: d[k], reverse=reverse)


def is_list_sorted(L, reverse=False):
    """
    Returns True if the list L appears to be sorted, otherwise False.

    Calling is_list_sorted(L,reverse=True) is the same as calling
    is_list_sorted(L.reverse(),reverse=False).

    Args:
        L (list): list to evaluate
        reverse (bool, optional): whether to reverse the list before evaluating sort status

    Returns:
        bool: True if the list L appears to be sorted, otherwise False
    """

    if reverse:
        return all(L[i] >= L[i + 1] for i in range(len(L) - 1))
    else:
        return all(L[i] <= L[i + 1] for i in range(len(L) - 1))


def generate_md_results_from_predictions_json(
    predictions_json_file, md_results_file, base_folder=None
):
    """
    Generate an MD-formatted .json file from a predictions.json file.  Typically,
    MD results files use relative paths, and predictions.json files use absolute paths, so
    this function optionally removes the leading string [base_folder] from all file names.

    Currently just applies the top classification category to every detection.  If the top classification
    is "blank", writes an empty detection list.

    speciesnet_to_md.py is a command-line driver for this function.

    Args:
        predictions_json_file (str): path to a predictions.json file
        md_results_file (str): path to which we should write an MD-formatted .json file
        base_folder (str, optional): leading string to remove from each path in the predictions.json file
    """

    # Read predictions file
    with open(predictions_json_file, "r") as f:
        predictions = json.load(f)
    predictions = predictions["predictions"]
    assert isinstance(predictions, list)

    # Convert backslashes to forward slashes in both filenames and the base folder string
    for im in predictions:
        im["filepath"] = im["filepath"].replace("\\", "/")
    if base_folder is not None:
        base_folder = base_folder.replace("\\", "/")

    detection_category_id_to_name = {}
    classification_category_name_to_id = {}

    # Keep track of detections that don't have an assigned detection category; these
    # are fake detections we create for non-blank images with non-empty detection lists.
    # We need to go back later and give them a legitimate detection category ID.
    all_unknown_detections = []

    # Create the output images list
    images_out = []

    # im_in = predictions[0]
    for im_in in predictions:

        # blank_prediction_string
        im_out = {}

        fn = im_in["filepath"]
        if base_folder is not None:
            if fn.startswith(base_folder):
                fn = fn.replace(base_folder, "", 1)

        im_out["file"] = fn

        if "failures" in im_in:

            im_out["failure"] = str(im_in["failures"])
            im_out["detections"] = None

        else:

            im_out["detections"] = []

            if "detections" in im_in:

                if len(im_in["detections"]) == 0:
                    im_out["detections"] = []
                else:
                    # det_in = im_in['detections'][0]
                    for det_in in im_in["detections"]:
                        det_out = {}
                        if det_in["category"] in detection_category_id_to_name:
                            assert (
                                detection_category_id_to_name[det_in["category"]]
                                == det_in["label"]
                            )
                        else:
                            detection_category_id_to_name[det_in["category"]] = det_in[
                                "label"
                            ]
                        det_out = {}
                        for s in ["category", "conf", "bbox"]:
                            det_out[s] = det_in[s]
                        im_out["detections"].append(det_out)

            # ...if detections are present

            class_to_assign = None
            class_confidence = None

            if "classifications" in im_in:

                classifications = im_in["classifications"]
                assert len(classifications["scores"]) == len(classifications["classes"])
                assert is_list_sorted(classifications["scores"], reverse=True)
                class_to_assign = classifications["classes"][0]
                class_confidence = classifications["scores"][0]

            if "prediction" in im_in:

                class_to_assign = im_in["prediction"]
                class_confidence = im_in["prediction_score"]

            if class_to_assign is not None:

                if class_to_assign == blank_prediction_string:

                    # This is a scenario that's not captured well by the MD format: a blank prediction
                    # with detections present.  But, for now, don't do anything special here, just making
                    # a note of this.
                    if len(im_out["detections"]) > 0:
                        pass

                else:

                    assert not class_to_assign.endswith("blank")

                    # This is a scenario that's not captured well by the MD format: no detections present,
                    # but a non-blank prediction.  For now, create a fake detection to handle this prediction.
                    if len(im_out["detections"]) == 0:

                        print(
                            "Warning: creating fake detection for non-blank whole-image classification"
                        )
                        det_out = {}
                        all_unknown_detections.append(det_out)

                        # We will change this to a string-int later
                        det_out["category"] = "unknown"
                        det_out["conf"] = class_confidence
                        det_out["bbox"] = [0, 0, 1, 1]
                        im_out["detections"].append(det_out)

                # ...if this is/isn't a blank classification

                # Attach that classification to each detection

                # Create a new category ID if necessary
                if class_to_assign in classification_category_name_to_id:
                    classification_category_id = classification_category_name_to_id[
                        class_to_assign
                    ]
                else:
                    classification_category_id = str(
                        len(classification_category_name_to_id)
                    )
                    classification_category_name_to_id[class_to_assign] = (
                        classification_category_id
                    )

                for det in im_out["detections"]:
                    det["classifications"] = []
                    det["classifications"].append(
                        [classification_category_id, class_confidence]
                    )

            # ...if we have some type of classification for this image

        # ...if this is/isn't a failure

        images_out.append(im_out)

    # ...for each image

    # Fix the 'unknown' category

    if len(all_unknown_detections) > 0:

        max_detection_category_id = max(
            [int(x) for x in detection_category_id_to_name.keys()]
        )
        unknown_category_id = str(max_detection_category_id + 1)
        detection_category_id_to_name[unknown_category_id] = "unknown"

        for det in all_unknown_detections:
            assert det["category"] == "unknown"
            det["category"] = unknown_category_id

    # Sort by filename

    images_out = sort_list_of_dicts_by_key(images_out, "file")

    # Prepare friendly classification names

    classification_category_descriptions = invert_dictionary(
        classification_category_name_to_id
    )
    classification_categories_out = {}
    for category_id in classification_category_descriptions.keys():
        category_name = classification_category_descriptions[category_id].split(";")[-1]
        classification_categories_out[category_id] = category_name

    # Prepare the output dict

    detection_categories_out = detection_category_id_to_name
    info = {}
    info["format_version"] = 1.4
    info["detector"] = "converted_from_predictions_json"

    output_dict = {}
    output_dict["info"] = info
    output_dict["detection_categories"] = detection_categories_out
    output_dict["classification_categories"] = classification_categories_out
    output_dict["classification_category_descriptions"] = (
        classification_category_descriptions
    )
    output_dict["images"] = images_out

    with open(md_results_file, "w") as f:
        json.dump(output_dict, f, indent=1)

    # TODO: ideally we would validate the output, but this requires a lot more imports,
    # so deferring this.
    """
    validation_options = ValidateBatchResultsOptions()
    validation_options.raise_errors = True    
    _ = validate_batch_results(md_results_file, options=validation_options)
    """


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "predictions_json_file",
        action="store",
        type=str,
        help=".json file to convert from SpeciesNet predictions.json format to MD format",
    )
    parser.add_argument(
        "md_results_file",
        action="store",
        type=str,
        help="output file to write in MD format",
    )
    parser.add_argument(
        "--base_folder",
        action="store",
        type=str,
        default=None,
        help="leading string to remove from each path in the predictions.json "
        + "file (to convert from absolute to relative paths)",
    )

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    generate_md_results_from_predictions_json(
        args.predictions_json_file, args.md_results_file, args.base_folder
    )


if __name__ == "__main__":
    main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to run the SpeciesNet model.

Provides a command-line interface to execute the SpeciesNet model on various inputs. It
uses flags for specifying input, output, and run options, allowing the user to run the
model in different modes.
"""

import json
import multiprocessing as mp
from pathlib import Path
from typing import Callable, Literal, Optional

from absl import app
from absl import flags

from speciesnet import DEFAULT_MODEL
from speciesnet import only_one_true
from speciesnet import SpeciesNet
from speciesnet.ensemble_prediction_combiner import PredictionType
from speciesnet.utils import load_partial_predictions
from speciesnet.utils import prepare_instances_dict

_MODEL = flags.DEFINE_string(
    "model",
    DEFAULT_MODEL,
    "SpeciesNet model to load.",
)
_CLASSIFIER_ONLY = flags.DEFINE_bool(
    "classifier_only",
    False,
    "Run only the classifier component. --classifier_only enables classifier-only mode, --noclassifier_only (default) disables it.",
)
_DETECTOR_ONLY = flags.DEFINE_bool(
    "detector_only",
    False,
    "Run only the detector component. --detector_only enables detector-only mode, --nodetector_only (default) disables it.",
)
_ENSEMBLE_ONLY = flags.DEFINE_bool(
    "ensemble_only",
    False,
    "Run only the ensemble component. --ensemble_only enables ensemble-only mode, --noensemble_only (default) disables it.",
)
_GEOFENCE = flags.DEFINE_bool(
    "geofence",
    True,
    "Enable geofencing during ensemble prediction. --geofence (default) enables geofencing, --nogeofence disables it.",
)
_INSTANCES_JSON = flags.DEFINE_string(
    "instances_json",
    None,
    "Input JSON file with instances to get predictions for.",
)
_FILEPATHS = flags.DEFINE_list(
    "filepaths",
    None,
    "List of image filepaths to get predictions for.",
)
_FILEPATHS_TXT = flags.DEFINE_string(
    "filepaths_txt",
    None,
    "Input TXT file with image filepaths to get predictions for.",
)
_FOLDERS = flags.DEFINE_list(
    "folders",
    None,
    "List of image folders to get predictions for.",
)
_FOLDERS_TXT = flags.DEFINE_string(
    "folders_txt",
    None,
    "Input TXT file with image folders to get predictions for.",
)
_COUNTRY = flags.DEFINE_string(
    "country",
    None,
    "Country (in ISO 3166-1 alpha-3 format, e.g. 'AUS') to enforce on all instances.",
)
_ADMIN1_REGION = flags.DEFINE_string(
    "admin1_region",
    None,
    "First-level administrative division (in ISO 3166-2 format, e.g. 'CA') to enforce on all "
    "instances.",
)
_TARGET_SPECIES_TXT = flags.DEFINE_string(
    "target_species_txt",
    None,
    "Input TXT file with species of interest to always compute classification scores for.",
)
_CLASSIFICATIONS_JSON = flags.DEFINE_string(
    "classifications_json",
    None,
    "Input JSON file with classifications from previous runs.",
)
_DETECTIONS_JSON = flags.DEFINE_string(
    "detections_json",
    None,
    "Input JSON file with detections from previous runs.",
)
_PREDICTIONS_JSON = flags.DEFINE_string(
    "predictions_json",
    None,
    "Output JSON file for storing computed predictions. If this file exists, only instances "
    "that are not already present in the output will be processed.",
)
_RUN_MODE = flags.DEFINE_enum(
    "run_mode",
    "multi_thread",
    ["multi_thread", "multi_process"],
    "Parallelism strategy.",
)
_BATCH_SIZE = flags.DEFINE_integer(
    "batch_size",
    8,
    "Batch size for classifier inference.",
)
_PROGRESS_BARS = flags.DEFINE_bool(
    "progress_bars",
    True,
    "Whether to show progress bars for the various inference components. --progress_bars "
    "(default) enables progress bars, --noprogress_bars disables them.",
)
_BYPASS_PROMPTS = flags.DEFINE_bool(
    "bypass_prompts",
    False,
    "Whether to bypass confirmation prompts when expected files aren't supplied, or "
    "unexpected files are supplied. --bypass_prompts bypasses prompts, --nobypass_prompts "
    "(default) does not.",
)


def guess_predictions_source(
    predictions: dict[str, dict],
) -> Literal["classifier", "detector", "ensemble", "unknown", "invalid"]:
    """Guesses which model component generated given predictions.

    Args:
        predictions: Dict of predictions, keyed by filepaths.

    Returns:
        Returns "classifier", "detector" or "ensemble" when the corresponding component
        was identified as the source of predictions. Returns "invalid" when predictions
        contain both classifications and detections, but couldn't identify results from
        the ensemble. Returns "unknown" when no prediction is recognizable (e.g. when
        there are only failures).
    """

    found_classifications = False
    found_detections = False
    found_ensemble_results = False

    for prediction in predictions.values():
        if "classifications" in prediction:
            found_classifications = True
        if "detections" in prediction:
            found_detections = True
        if "prediction" in prediction:
            found_ensemble_results = True
        if found_classifications and found_detections and not found_ensemble_results:
            return "invalid"

    if found_ensemble_results:
        return "ensemble"
    if found_classifications:
        return "classifier"
    if found_detections:
        return "detector"
    return "unknown"


def custom_combine_predictions_fn(
    *,
    classifications: dict[str, list],
    detections: list[dict],
    country: Optional[str],
    admin1_region: Optional[str],
    taxonomy_map: dict,
    geofence_map: dict,
    enable_geofence: bool,
    geofence_fn: Callable,
    roll_up_fn: Callable,
) -> PredictionType:
    """Ensembles classifications and detections in a custom way.

    Args:
        classifications:
            Dict of classification results. "classes" and "scores" are expected to be
            provided among the dict keys.
        detections:
            List of detection results, sorted in decreasing order of their confidence
            score. Each detection is expected to be a dict providing "label" and "conf"
            among its keys.
        country:
            Country (in ISO 3166-1 alpha-3 format) associated with predictions.
            Optional.
        admin1_region:
            First-level administrative division (in ISO 3166-2 format) associated with
            predictions. Optional.
        taxonomy_map:
            Dictionary mapping taxa to labels.
        geofence_map:
            Dictionary mapping full class strings to geofence rules.
        enable_geofence:
            Whether geofencing is enabled.
        geofence_fn:
            Callable to geofence animal classifications.
        roll_up_fn:
            Callable to roll up labels to the first matching level.

    Returns:
        A tuple of <label, score, prediction_source> describing the ensemble result.
    """

    del detections  # Unused.
    del country  # Unused.
    del admin1_region  # Unused.
    del taxonomy_map  # Unused.
    del geofence_map  # Unused.
    del enable_geofence  # Unused.
    del geofence_fn  # Unused.
    del roll_up_fn  # Unused.

    # Always return the second classifier prediction.
    return (
        classifications["classes"][1],
        classifications["scores"][1],
        "custom_ensemble",
    )


def say_yes_to_continue(question: str, stop_message: str) -> bool:
    if _BYPASS_PROMPTS.value:
        return True
    user_input = input(f"{question} [y/N]: ")
    if user_input.lower() in ["yes", "y"]:
        return True
    else:
        print(stop_message)
        return False


def local_file_exists(filepath: Optional[str]) -> bool:
    if not filepath:
        return False
    return Path(filepath).exists()


def main(argv: list[str]) -> None:
    del argv  # Unused.

    # Check for a valid combination of components to run.
    components = [_CLASSIFIER_ONLY, _DETECTOR_ONLY, _ENSEMBLE_ONLY]
    components_names = [f"--{c.name}" for c in components]
    components_values = [c.value for c in components]
    components_strings = [
        f"{name}={value}" for name, value in zip(components_names, components_values)
    ]
    if any(components_values) and not only_one_true(*components_values):
        raise ValueError(
            f"Expected at most one of [{', '.join(components_names)}] to be provided. "
            f"Received: [{', '.join(components_strings)}]."
        )
    if _ENSEMBLE_ONLY.value and (
        not _CLASSIFICATIONS_JSON.value or not _DETECTIONS_JSON.value
    ):
        raise ValueError(
            f"Expected --{_CLASSIFICATIONS_JSON.name} and --{_DETECTIONS_JSON.name} to "
            f"be set when --{_ENSEMBLE_ONLY.name} is requested."
        )
    if _CLASSIFIER_ONLY.value:
        components = "classifier"
    elif _DETECTOR_ONLY.value:
        components = "detector"
    elif _ENSEMBLE_ONLY.value:
        components = "ensemble"
    else:
        components = "all"

    # Check for valid inputs.
    inputs = [_INSTANCES_JSON, _FILEPATHS, _FILEPATHS_TXT, _FOLDERS, _FOLDERS_TXT]
    inputs_names = [f"--{i.name}" for i in inputs]
    inputs_values = [i.value for i in inputs]
    inputs_strings = [
        f"{name}={value}" for name, value in zip(inputs_names, inputs_values)
    ]
    if not only_one_true(*inputs_values):
        raise ValueError(
            f"Expected exactly one of [{', '.join(inputs_names)}] to be provided. "
            f"Received: [{', '.join(inputs_strings)}]."
        )
    instances_dict = prepare_instances_dict(
        instances_json=_INSTANCES_JSON.value,
        filepaths=list(_FILEPATHS.value) if _FILEPATHS.value else None,
        filepaths_txt=_FILEPATHS_TXT.value,
        folders=list(_FOLDERS.value) if _FOLDERS.value else None,
        folders_txt=_FOLDERS_TXT.value,
        country=_COUNTRY.value,
        admin1_region=_ADMIN1_REGION.value,
    )

    # Check the compatibility of output predictions with existing partial predictions.
    if _PREDICTIONS_JSON.value:
        partial_predictions, _ = load_partial_predictions(
            _PREDICTIONS_JSON.value, instances_dict["instances"]
        )
        predictions_source = guess_predictions_source(partial_predictions)

        if _CLASSIFIER_ONLY.value and predictions_source not in [
            "classifier",
            "unknown",
        ]:
            raise RuntimeError(
                f"The classifier risks overwriting previous predictions from "
                f"`{_PREDICTIONS_JSON.value}` that were produced by different "
                f"components. Make sure to provide a different output location to "
                f"--{_PREDICTIONS_JSON.name}."
            )

        if _DETECTOR_ONLY.value and predictions_source not in ["detector", "unknown"]:
            raise RuntimeError(
                f"The detector risks overwriting previous predictions from "
                f"`{_PREDICTIONS_JSON.value}` that were produced by different "
                f"components. Make sure to provide a different output location to "
                f"--{_PREDICTIONS_JSON.name}."
            )

        if _ENSEMBLE_ONLY.value and predictions_source not in ["ensemble", "unknown"]:
            raise RuntimeError(
                f"The ensemble risks overwriting previous predictions from "
                f"`{_PREDICTIONS_JSON.value}` that were produced by different "
                f"components. Make sure to provide a different output location to "
                f"--{_PREDICTIONS_JSON.name}."
            )

    else:
        if not say_yes_to_continue(
            question="Continue without saving predictions to a JSON file?",
            stop_message=(
                f"Please provide an output filepath via --{_PREDICTIONS_JSON.name}."
            ),
        ):
            return

    # If a list of target species is given, check that it exists
    if _TARGET_SPECIES_TXT.value is not None and not local_file_exists(
        _TARGET_SPECIES_TXT.value
    ):
        raise RuntimeError(
            f"Target species file '{_TARGET_SPECIES_TXT.value}' specified via --{_PREDICTIONS_JSON.name} does not exist."
        )

    # Load classifications and/or detections from previous runs.
    classifications_dict, _ = load_partial_predictions(
        _CLASSIFICATIONS_JSON.value, instances_dict["instances"]
    )
    detections_dict, _ = load_partial_predictions(
        _DETECTIONS_JSON.value, instances_dict["instances"]
    )

    # Set running mode.
    run_mode: Literal["multi_thread", "multi_process"] = _RUN_MODE.value  # type: ignore
    mp.set_start_method("spawn")

    # Make predictions.
    model = SpeciesNet(
        _MODEL.value,
        components=components,
        geofence=_GEOFENCE.value,
        target_species_txt=_TARGET_SPECIES_TXT.value,
        # Uncomment the line below if you want to run your own custom ensembling
        # routine. And also, implement that routine! :-)
        # combine_predictions_fn=custom_combine_predictions_fn,
        multiprocessing=(run_mode == "multi_process"),
    )
    if hasattr(model, "classifier") and not hasattr(model, "detector"):
        if (
            model.classifier.model_info.type_ == "always_crop"
            and not local_file_exists(_DETECTIONS_JSON.value)
        ):
            if not say_yes_to_continue(
                question=(
                    "Classifier expects detections JSON. Continue without providing "
                    "this file and run classifier on full images instead of crops?"
                ),
                stop_message=(
                    f"Please provide detections via --{_DETECTIONS_JSON.name} and make "
                    "sure that file exists."
                ),
            ):
                return
        elif (
            model.classifier.model_info.type_ == "full_image" and _DETECTIONS_JSON.value
        ):
            if not say_yes_to_continue(
                question=(
                    "Classifier doesn't expect detections JSON, yet such file was "
                    f"provided via --{_DETECTIONS_JSON.name}. Continue even though "
                    "given detections JSON will have no effect?"
                ),
                stop_message=f"Please drop the --{_DETECTIONS_JSON.name} flag.",
            ):
                return
    if _CLASSIFIER_ONLY.value:
        predictions_dict = model.classify(
            instances_dict=instances_dict,
            detections_dict=detections_dict,
            run_mode=run_mode,
            batch_size=_BATCH_SIZE.value,
            progress_bars=_PROGRESS_BARS.value,
            predictions_json=_PREDICTIONS_JSON.value,
        )
    elif _DETECTOR_ONLY.value:
        predictions_dict = model.detect(
            instances_dict=instances_dict,
            run_mode=run_mode,
            progress_bars=_PROGRESS_BARS.value,
            predictions_json=_PREDICTIONS_JSON.value,
        )
    elif _ENSEMBLE_ONLY.value:
        predictions_dict = model.ensemble_from_past_runs(
            instances_dict=instances_dict,
            classifications_dict=classifications_dict,
            detections_dict=detections_dict,
            progress_bars=_PROGRESS_BARS.value,
            predictions_json=_PREDICTIONS_JSON.value,
        )
    else:
        predictions_dict = model.predict(
            instances_dict=instances_dict,
            run_mode=run_mode,
            batch_size=_BATCH_SIZE.value,
            progress_bars=_PROGRESS_BARS.value,
            predictions_json=_PREDICTIONS_JSON.value,
        )
    if predictions_dict is not None:
        print(
            "Predictions:\n"
            + json.dumps(predictions_dict, ensure_ascii=False, indent=4)
        )


if __name__ == "__main__":
    app.run(main)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-module-docstring
# pylint: disable=protected-access

import warnings

import pytest

from speciesnet.constants import Classification
from speciesnet.ensemble import PredictionType
from speciesnet.ensemble import SpeciesNetEnsemble
from speciesnet.taxonomy_utils import get_ancestor_at_level
from speciesnet.utils import load_rgb_image

# fmt: off
# pylint: disable=line-too-long

BLANK = "f1856211-cfb7-4a5b-9158-c0f72fd09ee6;;;;;;blank"
BLANK_FC = ";;;;"
HUMAN = "990ae9dd-7a59-4344-afcb-1b7b21368000;mammalia;primates;hominidae;homo;sapiens;human"
HUMAN_FC = "mammalia;primates;hominidae;homo;sapiens"
VEHICLE = "e2895ed5-780b-48f6-8a11-9e27cb594511;;;;;;vehicle"
VEHICLE_FC = ";;;;"

LION = "ddf59264-185a-4d35-b647-2785792bdf54;mammalia;carnivora;felidae;panthera;leo;lion"
LION_FC = "mammalia;carnivora;felidae;panthera;leo"
PANTHERA_GENUS = "fbb23d07-6677-43db-b650-f99ac452c50f;mammalia;carnivora;felidae;panthera;;panthera species"
PANTHERA_GENUS_FC = "mammalia;carnivora;felidae;panthera;"
FELIDAE_FAMILY = "df8514b0-10a5-411f-8ed6-0f415e8153a3;mammalia;carnivora;felidae;;;cat family"
FELIDAE_FAMILY_FC = "mammalia;carnivora;felidae;;"
CARNIVORA_ORDER = "eeeb5d26-2a47-4d01-a3de-10b33ec0aee4;mammalia;carnivora;;;;carnivorous mammal"
CARNIVORA_ORDER_FC = "mammalia;carnivora;;;"
MAMMALIA_CLASS = "f2d233e3-80e3-433d-9687-e29ecc7a467a;mammalia;;;;;mammal"
MAMMALIA_CLASS_FC = "mammalia;;;;"
ANIMAL_KINGDOM = "1f689929-883d-4dae-958c-3d57ab5b6c16;;;;;;animal"
ANIMAL_KINGDOM_FC = ";;;;"

BROWN_BEAR = "330bb1e9-84d6-4e41-afa9-938aee17ea29;mammalia;carnivora;ursidae;ursus;arctos;brown bear"
BROWN_BEAR_FC = "mammalia;carnivora;ursidae;ursus;arctos"
POLAR_BEAR = "e7f83bf6-df2c-4ce0-97fc-2f233df23ec4;mammalia;carnivora;ursidae;ursus;maritimus;polar bear"
POLAR_BEAR_FC = "mammalia;carnivora;ursidae;ursus;maritimus"
GIANT_PANDA = "85662682-67c1-4ecb-ba05-ba12e2df6b65;mammalia;carnivora;ursidae;ailuropoda;melanoleuca;giant panda"
GIANT_PANDA_FC = "mammalia;carnivora;ursidae;ailuropoda;melanoleuca"
URSUS_GENUS = "5a0f5e3f-c634-4b86-910a-b105cb526a24;mammalia;carnivora;ursidae;ursus;;ursus species"
URSUS_GENUS_FC = "mammalia;carnivora;ursidae;ursus;"
URSIDAE_FAMILY = "ec1a70f4-41c0-4aba-9150-292fb2b7a324;mammalia;carnivora;ursidae;;;bear family"
URSIDAE_FAMILY_FC = "mammalia;carnivora;ursidae;;"

PUMA = "9c564562-9429-405c-8529-04cff7752282;mammalia;carnivora;felidae;puma;concolor;puma"
PUMA_FC = "mammalia;carnivora;felidae;puma;concolor"
SAND_CAT = "e588253d-d61d-4149-a96c-8c245927a80f;mammalia;carnivora;felidae;felis;margarita;sand cat"
SAND_CAT_FC = "mammalia;carnivora;felidae;felis;margarita"

# pylint: enable=line-too-long
# fmt: on


class TestEnsemble:
    """Tests for the ensemble component."""

    @pytest.fixture(scope="class")
    def ensemble(self, model_name: str) -> SpeciesNetEnsemble:
        return SpeciesNetEnsemble(model_name)

    @pytest.fixture
    def mock_ensemble(self, monkeypatch, ensemble) -> SpeciesNetEnsemble:
        taxonomy_map = {
            BLANK_FC: BLANK,
            HUMAN_FC: HUMAN,
            VEHICLE_FC: VEHICLE,
            LION_FC: LION,
            PANTHERA_GENUS_FC: PANTHERA_GENUS,
            FELIDAE_FAMILY_FC: FELIDAE_FAMILY,
            CARNIVORA_ORDER_FC: CARNIVORA_ORDER,
            MAMMALIA_CLASS_FC: MAMMALIA_CLASS,
            ANIMAL_KINGDOM_FC: ANIMAL_KINGDOM,
            BROWN_BEAR_FC: BROWN_BEAR,
            POLAR_BEAR_FC: POLAR_BEAR,
            GIANT_PANDA_FC: GIANT_PANDA,
            URSUS_GENUS_FC: URSUS_GENUS,
            URSIDAE_FAMILY_FC: URSIDAE_FAMILY,
        }
        monkeypatch.setattr(ensemble, "taxonomy_map", taxonomy_map)

        geofence_map = {
            LION_FC: {
                "allow": {
                    "KEN": [],
                    "TZA": [],
                }
            },
            PANTHERA_GENUS_FC: {
                "allow": {
                    "KEN": [],
                    "TZA": [],
                    "USA": ["AK", "CA"],
                }
            },
            FELIDAE_FAMILY_FC: {
                "allow": {
                    "FRA": [],
                    "KEN": [],
                    "TZA": [],
                    "USA": [],
                },
                "block": {
                    "FRA": [],
                    "USA": ["NY"],
                },
            },
            SAND_CAT_FC: {
                "block": {
                    "AUS": [],
                },
            },
            URSIDAE_FAMILY_FC: {
                "block": {
                    "GBR": [],
                },
            },
        }
        monkeypatch.setattr(ensemble, "geofence_map", geofence_map)

        return ensemble

    @pytest.fixture
    def mock_ensemble_no_geofence(
        self, monkeypatch, mock_ensemble
    ) -> SpeciesNetEnsemble:
        monkeypatch.setattr(mock_ensemble, "enable_geofence", False)
        return mock_ensemble

    @pytest.fixture
    def mock_ensemble2(self, monkeypatch, ensemble) -> SpeciesNetEnsemble:

        def prediction_combiner_mock(
            classifications: dict[str, list], *args, **kwargs
        ) -> PredictionType:
            del args  # Unused.
            del kwargs  # Unused.
            return classifications["classes"][0], classifications["scores"][0], "mock"

        monkeypatch.setattr(
            ensemble,
            "prediction_combiner",
            prediction_combiner_mock,
        )
        return ensemble

    def test_combine(self, mock_ensemble2) -> None:

        blank_img = load_rgb_image("test_data/blank.jpg")
        blank2_img = load_rgb_image("test_data/blank2.jpg")
        blank3_img = load_rgb_image("test_data/blank3.jpg")
        assert blank_img and blank2_img and blank3_img

        expected_model_version = mock_ensemble2.model_info.version

        filepaths = [
            "a.jpg",
            "b.jpg",
            "c.jpg",
            "d.jpg",
            "e.jpg",
            "f.jpg",
        ]
        classifier_results = {
            "b.jpg": {
                "filepath": "b.jpg",
                "failures": ["CLASSIFIER"],
            },
            "c.jpg": {
                "filepath": "c.jpg",
                "classifications": {
                    "classes": ["X", "Y", "Z"],
                    "scores": [0.5, 0.3, 0.2],
                },
            },
            "d.jpg": {
                "filepath": "d.jpg",
                "classifications": {
                    "classes": ["R", "S", "T"],
                    "scores": [0.7, 0.2, 0.1],
                },
            },
            "e.jpg": {
                "filepath": "e.jpg",
                "classifications": {
                    "classes": ["K", "L", "M"],
                    "scores": [0.9, 0.1, 0.0],
                },
            },
            "f.jpg": {
                "filepath": "f.jpg",
                "classifications": {
                    "classes": ["K", "L", "M"],
                    "scores": [0.9, 0.1, 0.0],
                },
            },
        }
        detector_results = {
            "a.jpg": {
                "filepath": "a.jpg",
                "failures": ["DETECTOR"],
            },
            "b.jpg": {
                "filepath": "b.jpg",
                "detections": [
                    {
                        "category": "1",
                        "label": "animal",
                        "conf": 0.5,
                        "bbox": [0.0, 0.1, 0.2, 0.3],
                    }
                ],
            },
            "d.jpg": {
                "filepath": "d.jpg",
                "detections": [
                    {
                        "category": "2",
                        "label": "human",
                        "conf": 0.7,
                        "bbox": [0.1, 0.2, 0.3, 0.4],
                    }
                ],
            },
            "e.jpg": {
                "filepath": "e.jpg",
                "detections": [
                    {
                        "category": "2",
                        "label": "human",
                        "conf": 0.7,
                        "bbox": [0.1, 0.2, 0.3, 0.4],
                    }
                ],
            },
            "f.jpg": {
                "filepath": "f.jpg",
                "detections": [],
            },
        }
        geolocation_results = {
            "a.jpg": {
                "country": "COUNTRY_A",
            },
            "b.jpg": {
                "country": "COUNTRY_B",
            },
            "c.jpg": {
                "country": "COUNTRY_C",
            },
            "e.jpg": {
                "country": "COUNTRY_E",
            },
            "f.jpg": {
                "country": "COUNTRY_F",
            },
        }
        partial_predictions = {
            "f.jpg": {
                "filepath": "f.jpg",
                "classifications": {
                    "classes": ["XYZ"],
                    "scores": [0.8],
                },
                "detections": [],
                "prediction": "XYZ",
                "prediction_score": 0.4,
                "prediction_source": "partial",
                "model_version": expected_model_version,
            },
        }

        assert mock_ensemble2.combine(
            filepaths,
            classifier_results,
            detector_results,
            geolocation_results,
            partial_predictions,
        ) == [
            {
                "filepath": "a.jpg",
                "failures": ["CLASSIFIER", "DETECTOR"],
                "country": "COUNTRY_A",
                "model_version": expected_model_version,
            },
            {
                "filepath": "b.jpg",
                "failures": ["CLASSIFIER"],
                "country": "COUNTRY_B",
                "detections": [
                    {
                        "category": "1",
                        "label": "animal",
                        "conf": 0.5,
                        "bbox": [0.0, 0.1, 0.2, 0.3],
                    }
                ],
                "model_version": expected_model_version,
            },
            {
                "filepath": "c.jpg",
                "failures": ["DETECTOR"],
                "country": "COUNTRY_C",
                "classifications": {
                    "classes": ["X", "Y", "Z"],
                    "scores": [0.5, 0.3, 0.2],
                },
                "model_version": expected_model_version,
            },
            {
                "filepath": "d.jpg",
                "failures": ["GEOLOCATION"],
                "classifications": {
                    "classes": ["R", "S", "T"],
                    "scores": [0.7, 0.2, 0.1],
                },
                "detections": [
                    {
                        "category": "2",
                        "label": "human",
                        "conf": 0.7,
                        "bbox": [0.1, 0.2, 0.3, 0.4],
                    }
                ],
                "prediction": "R",
                "prediction_score": 0.7,
                "prediction_source": "mock",
                "model_version": expected_model_version,
            },
            {
                "filepath": "e.jpg",
                "country": "COUNTRY_E",
                "classifications": {
                    "classes": ["K", "L", "M"],
                    "scores": [0.9, 0.1, 0.0],
                },
                "detections": [
                    {
                        "category": "2",
                        "label": "human",
                        "conf": 0.7,
                        "bbox": [0.1, 0.2, 0.3, 0.4],
                    }
                ],
                "prediction": "K",
                "prediction_score": 0.9,
                "prediction_source": "mock",
                "model_version": expected_model_version,
            },
            {
                "filepath": "f.jpg",
                "classifications": {
                    "classes": ["XYZ"],
                    "scores": [0.8],
                },
                "detections": [],
                "prediction": "XYZ",
                "prediction_score": 0.4,
                "prediction_source": "partial",
                "model_version": expected_model_version,
            },
        ]

    def test_complete_taxonomy(self, ensemble) -> None:

        missing_ancestors = set()

        taxonomy_levels = {
            "kingdom": 0,
            "class": 1,
            "order": 2,
            "family": 3,
            "genus": 4,
            "species": 5,
        }

        def _taxa_from_label(label: str) -> str:
            return ";".join(label.split(";")[1:6])

        def _level_idx_from_label(label: str) -> int:
            label_parts = label.split(";")
            for idx in range(5, 0, -1):
                if label_parts[idx]:
                    return idx
            return 0

        def _ancestor_from_label(label: str, level_idx: int) -> str:
            label_parts = label.split(";")
            return ";".join(label_parts[1 : level_idx + 1]) + (";" * (5 - level_idx))

        with open(
            ensemble.model_info.classifier_labels, mode="r", encoding="utf-8"
        ) as fp:
            classifier_labels = [line.strip() for line in fp.readlines()]

        for label in [Classification.HUMAN, Classification.ANIMAL]:
            taxa = _taxa_from_label(label)
            assert taxa in ensemble.taxonomy_map
            assert ensemble.taxonomy_map[taxa] == label
        for label in [Classification.UNKNOWN]:
            taxa = _taxa_from_label(label)
            assert taxa not in ensemble.taxonomy_map

        for label in classifier_labels:
            if label in [
                Classification.BLANK,
                Classification.VEHICLE,
                Classification.UNKNOWN,
            ]:
                continue

            max_level_idx = _level_idx_from_label(label)

            for taxonomy_level, taxonomy_level_idx in taxonomy_levels.items():
                ancestor = get_ancestor_at_level(
                    label, taxonomy_level, ensemble.taxonomy_map
                )
                if taxonomy_level_idx <= max_level_idx:
                    if not ancestor:
                        missing_ancestors.add(
                            _ancestor_from_label(label, taxonomy_level_idx)
                        )
                else:
                    assert ancestor is None

        if missing_ancestors:
            warnings.warn(
                UserWarning(
                    "Missing from taxonomy: \n" + "\n".join(sorted(missing_ancestors))
                )
            )

    def test_complete_geofence(self, ensemble) -> None:

        missing_ancestors = set()

        def _taxa_from_label(label: str) -> str:
            return ";".join(label.split(";")[1:6])

        def _ancestor_from_label(label: str, level_idx: int) -> str:
            label_parts = label.split(";")
            return ";".join(label_parts[1 : level_idx + 1]) + (";" * (5 - level_idx))

        with open(
            ensemble.model_info.classifier_labels, mode="r", encoding="utf-8"
        ) as fp:
            classifier_labels = [line.strip() for line in fp.readlines()]

        for label in [
            Classification.BLANK,
            Classification.ANIMAL,
            Classification.VEHICLE,
            Classification.UNKNOWN,
        ]:
            taxa = _taxa_from_label(label)
            assert taxa not in ensemble.geofence_map

        for label in [Classification.HUMAN]:
            taxa = _taxa_from_label(label)
            assert taxa in ensemble.geofence_map

        for label in classifier_labels:
            if label in [
                Classification.BLANK,
                Classification.ANIMAL,
                Classification.VEHICLE,
                Classification.UNKNOWN,
            ]:
                continue

            for level_idx in range(1, 6):
                ancestor = _ancestor_from_label(label, level_idx)
                if ancestor not in ensemble.geofence_map:
                    missing_ancestors.add(ancestor)

        if missing_ancestors:
            warnings.warn(
                UserWarning(
                    "Missing from geofence: \n" + "\n".join(sorted(missing_ancestors))
                )
            )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script to run the SpeciesNet server.

Sets up and runs an HTTP server using the `litserve` library, exposing the
SpeciesNet model for remote inference. It provides a REST API for making
prediction requests to the model.
"""

from typing import Optional

from absl import app
from absl import flags
from fastapi import HTTPException
import litserve as ls

from speciesnet import DEFAULT_MODEL
from speciesnet import file_exists
from speciesnet import SpeciesNet

_PORT = flags.DEFINE_integer(
    "port",
    8000,
    "Port to run the server on.",
)
_API_PATH = flags.DEFINE_string(
    "api_path",
    "/predict",
    "URL path for the server endpoint.",
)
_WORKERS_PER_DEVICE = flags.DEFINE_integer(
    "workers_per_device",
    1,
    "Number of server replicas per device.",
)
_TIMEOUT = flags.DEFINE_integer(
    "timeout",
    30,
    "Timeout (in seconds) for requests.",
)
_BACKLOG = flags.DEFINE_integer(
    "backlog",
    2048,
    "Maximum number of connections to hold in backlog.",
)
_MODEL = flags.DEFINE_string(
    "model",
    DEFAULT_MODEL,
    "SpeciesNet model to load.",
)
_GEOFENCE = flags.DEFINE_bool(
    "geofence",
    True,
    "Whether to enable geofencing or not.",
)
_EXTRA_FIELDS = flags.DEFINE_list(
    "extra_fields",
    None,
    "Comma-separated list of extra fields to propagate from request to response.",
)


class SpeciesNetLitAPI(ls.LitAPI):
    """Core API to serve the SpeciesNet model.

    This class implements the server side of SpeciesNet by implementing LitAPI interface
    required by the `litserve` library. It handles request parsing, model loading,
    inference, and response formatting. This is a bridge between HTTP requests and the
    internal Python API for SpeciesNet.
    """

    def __init__(
        self,
        model_name: str,
        geofence: bool = True,
        extra_fields: Optional[list[str]] = None,
    ) -> None:
        """Initializes the SpeciesNet API server.

        Args:
            model_name:
                String value identifying the model to be loaded. It can be a Kaggle
                identifier (starting with `kaggle:`), a HuggingFace identifier (starting
                with `hf:`) or a local folder to load the model from.
            geofence:
                Whether to enable geofencing or not. Defaults to `True`.
            extra_fields:
                 Comma-separated list of extra fields to propagate from request to
                 response.
        """
        super().__init__()
        self.model_name = model_name
        self.geofence = geofence
        self.extra_fields = extra_fields or []

    def setup(self, device):
        del device  # Unused.
        self.model = SpeciesNet(self.model_name, geofence=self.geofence)

    def decode_request(self, request, context):
        del context  # Unused.
        for instance in request["instances"]:
            filepath = instance["filepath"]
            if not file_exists(filepath):
                raise HTTPException(400, f"Cannot access filepath: `{filepath}`")
        return request

    def _propagate_extra_fields(
        self, instances_dict: dict, predictions_dict: dict
    ) -> dict:
        predictions = predictions_dict["predictions"]
        new_predictions = {p["filepath"]: p for p in predictions}
        for instance in instances_dict["instances"]:
            for field in self.extra_fields:
                if field in instance:
                    new_predictions[instance["filepath"]][field] = instance[field]
        return {"predictions": list(new_predictions.values())}

    def predict(self, instances_dict, context):
        del context  # Unused.
        predictions_dict = self.model.predict(instances_dict=instances_dict)
        assert predictions_dict is not None
        return self._propagate_extra_fields(instances_dict, predictions_dict)

    def encode_response(self, output, context):
        del context  # Unused.
        return output


def main(argv: list[str]) -> None:
    del argv  # Unused.

    api = SpeciesNetLitAPI(
        model_name=_MODEL.value,
        geofence=_GEOFENCE.value,
        extra_fields=_EXTRA_FIELDS.value,
    )
    server = ls.LitServer(
        api,
        accelerator="auto",
        devices="auto",
        workers_per_device=_WORKERS_PER_DEVICE.value,
        timeout=_TIMEOUT.value,
        api_path=_API_PATH.value,
    )
    server.run(
        port=_PORT.value,
        generate_client_file=False,
        backlog=_BACKLOG.value,
    )


if __name__ == "__main__":
    app.run(main)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multiprocessing utilities to run SpeciesNet.

Provides utilities for running the SpeciesNet model with various forms of
parallelization, including multi-threading and multiprocessing. It implements a main
`SpeciesNet` class, that serves as a high level interface to interact with the different
SpeciesNet components using different parallelization strategies.
"""

__all__ = [
    "SpeciesNet",
]

import multiprocessing as mp
from multiprocessing.managers import SyncManager
from multiprocessing.pool import ThreadPool
from pathlib import Path
import queue
import threading
import traceback
from typing import Callable, Literal, Optional, Union

from absl import logging
from tqdm import tqdm

from speciesnet.classifier import SpeciesNetClassifier
from speciesnet.constants import Failure
from speciesnet.detector import SpeciesNetDetector
from speciesnet.ensemble import SpeciesNetEnsemble
from speciesnet.ensemble_prediction_combiner import combine_predictions_for_single_item
from speciesnet.geolocation import find_admin1_region
from speciesnet.utils import BBox
from speciesnet.utils import load_partial_predictions
from speciesnet.utils import load_rgb_image
from speciesnet.utils import prepare_instances_dict
from speciesnet.utils import PreprocessedImage
from speciesnet.utils import save_predictions

# Handy type aliases.
StrPath = Union[str, Path]
DetectorInput = tuple[str, Optional[PreprocessedImage]]
BBoxOutput = tuple[str, list[BBox]]
ClassifierInput = tuple[str, Optional[PreprocessedImage]]

# Register SpeciesNet model components with the SyncManager to be able to safely share
# them between processes.
SyncManager.register("Classifier", SpeciesNetClassifier)
SyncManager.register("Detector", SpeciesNetDetector)
SyncManager.register("Ensemble", SpeciesNetEnsemble)


class RepeatedAction(threading.Thread):
    """Repeated action to run regularly at a given interval.

    Implements a threading mechanism to execute a specific function repeatedly at set
    time intervals. It's commonly used for background tasks such as saving partial
    results periodically during long-running inference jobs.
    """

    def __init__(self, interval: float, fn: Callable, *fn_args, **fn_kwargs) -> None:
        """Initializes the repeated action.

        Args:
            interval:
                Number of seconds (can be fractional) to wait before repeating the
                action.
            fn:
                Callable representing the action to repeat.
            *fn_args:
                Arguments for the action to repeat.
            **fn_kwargs:
                Keyword arguments for the action to repeat.
        """

        super().__init__()
        self._interval = interval
        self._fn = fn
        self._fn_args = fn_args
        self._fn_kwargs = fn_kwargs
        self._stopped = threading.Event()

    def run(self) -> None:
        """Starts the repeated action."""

        while not self._stopped.is_set():
            self._stopped.wait(self._interval)
            if self._stopped.is_set():
                break
            self._fn(*self._fn_args, **self._fn_kwargs)

    def stop(self) -> None:
        """Stops the repeated action."""

        self._stopped.set()


class Progress:
    """Progress tracker for different components of SpeciesNet.

    Provides a mechanism to track the progress of various tasks within the SpeciesNet
    inference process. It uses `tqdm` progress bars to visually show the status of each
    component, like detector preprocessing, detector prediction, classifier
    preprocessing, classifier prediction, and geolocation operations. It offers a way to
    update individual trackers as the inference progresses, and to stop tracking when
    inference is complete.
    """

    def __init__(
        self,
        enabled: list[str],
        total: Optional[int] = None,
        batches: Optional[int] = None,
        rlock: Optional[threading.RLock] = None,
    ) -> None:
        """Initializes the progress tracker.

        Args:
            enabled:
                List of enabled trackers from the following list:
                ["detector_preprocess", "detector_predict", "classifier_preprocess",
                "classifier_predict", "geolocation"].
            total:
                Number of expected iterations. Optional.
            batches:
                Number of expected batches. Optional.
            rlock:
                RLock object to use as the global tracking lock. Optional.
        """

        tqdm.monitor_interval = 0
        if rlock:
            tqdm.set_lock(rlock)

        self._pbars = {}
        if "detector_preprocess" in enabled:
            self._pbars["detector_preprocess"] = tqdm(
                desc="Detector preprocess   ",
                total=total,
                mininterval=0,
                colour="#029e73",
            )
        if "detector_predict" in enabled:
            self._pbars["detector_predict"] = tqdm(
                desc="Detector predict      ",
                total=total,
                mininterval=0,
                colour="#ece133",
            )
        if "classifier_preprocess" in enabled:
            self._pbars["classifier_preprocess"] = tqdm(
                desc="Classifier preprocess ",
                total=total,
                mininterval=0,
                colour="#d55e00",
            )
        if "classifier_predict" in enabled:
            self._pbars["classifier_predict"] = tqdm(
                desc="Classifier predict    ",
                total=batches or total,
                mininterval=0,
                colour="#0184cb",
            )
        if "geolocation" in enabled:
            self._pbars["geolocation"] = tqdm(
                desc="Geolocation           ",
                total=total,
                mininterval=0,
                colour="#01b2b2",
            )

    def update(self, name: str) -> None:
        """Updates individual tracker.

        Args:
            name:
                Name of the individual tracker to update.
        """

        if name in self._pbars:
            self._pbars[name].update()

    def stop(self) -> None:
        """Stops progress tracking."""

        for pbar in self._pbars.values():
            pbar.close()


def _prepare_detector_input(
    detector: SpeciesNetDetector,
    filepath: str,  # input
    detector_queue: queue.Queue[DetectorInput],  # output
) -> None:
    """Prepares the input for detector inference.

    Responsible for loading and preprocessing an image in preparation for the detector.

    Args:
        detector:
            SpeciesNetDetector to use.
        filepath:
            Path to image to load and preprocess.
        detector_queue:
            Output queue for preprocessed images for detector inference.
    """

    img = load_rgb_image(filepath)
    try:
        img = detector.preprocess(img)
        detector_queue.put((filepath, img))
    except:
        detector_queue.put((filepath, None))
        raise


def _run_detector(
    detector: SpeciesNetDetector,
    input_queue: queue.Queue[DetectorInput],  # input
    results_dict: dict,  # output
    bboxes_queue: Optional[queue.Queue[BBoxOutput]] = None,  # output
) -> None:
    """Runs detector inference.

    Takes a preprocessed image from the input queue and runs the detector model on it.
    The raw output of the detector is stored in `results_dict` and the list of bounding
    boxes is additionally stored in `bboxes_queue`.

    Args:
        detector:
            SpeciesNetDetector to run.
        input_queue:
            Input queue of preprocessed images.
        results_dict:
            Output dict for inference results.
        bboxes_queue:
            Output queue for bounding boxes identified during inference. Optional.
    """

    filepath, img = input_queue.get()
    prediction = detector.predict(filepath, img)
    results_dict[filepath] = prediction
    if bboxes_queue:
        detections = prediction.get("detections", [])
        bboxes_queue.put((filepath, [BBox(*det["bbox"]) for det in detections]))


def _prepare_classifier_input(
    classifier: SpeciesNetClassifier,
    bboxes_queue: queue.Queue[BBoxOutput],  # input
    classifier_queue: queue.Queue[ClassifierInput],  # output
) -> None:
    """Prepares the input for classifier inference.

    Takes bounding box information from `bboxes_queue` and uses it to load and
    preprocess the image for classifier inference. It first loads the image,
    preprocesses it by potentially cropping based on bboxes and finally outputs the
    image into `classifier_queue` to be used by the classifier model.

    Args:
        classifier:
            SpeciesNetClassifier to use.
        bboxes_queue:
            Input queue of bounding boxes from detector inference.
        classifier_queue:
            Output queue for preprocessed images for classifier inference.
    """

    filepath, bboxes = bboxes_queue.get()
    img = load_rgb_image(filepath)
    try:
        img = classifier.preprocess(img, bboxes=bboxes)
        classifier_queue.put((filepath, img))
    except:
        classifier_queue.put((filepath, None))
        raise


def _run_classifier(
    classifier: SpeciesNetClassifier,
    input_queue: queue.Queue[ClassifierInput],  # input
    results_dict: dict,  # output
    batch_size: int,
) -> None:
    """Runs classifier inference.

    Takes a batch of preprocessed images from the input queue and runs the classifier
    model on it. The output of the classifier is stored in `results_dict`.

    Args:
        classifier:
            SpeciesNetClassifier to run.
        input_queue:
            Input queue of preprocessed images.
        results_dict:
            Output dict for inference results.
        batch_size:
            Batch size for inference.

    """

    input_tuples = [input_queue.get() for _ in range(batch_size)]
    filepaths = [t[0] for t in input_tuples]
    imgs = [t[1] for t in input_tuples]
    predictions = classifier.batch_predict(filepaths, imgs)
    for filepath, prediction in zip(filepaths, predictions):
        results_dict[filepath] = prediction


def _find_admin1_region(  # pylint: disable=too-many-positional-arguments
    filepath: str,  # input
    country: Optional[str],  # input
    admin1_region: Optional[str],  # input
    latitude: Optional[float],  # input
    longitude: Optional[float],  # input
    results_dict: dict,  # output
) -> None:
    """Find the first-level administrative division for a given (lat, lon) location.

    This function uses the provided geographic information to find the first-level
    administrative division (e.g., state in the USA) using the `find_admin1_region()`
    function. The result is stored in the `results_dict`.

    Args:
        filepath:
            Image filepath.
        country:
            Country in ISO 3166-1 alpha-3 format. Optional.
        admin1_region:
            First-level administrative division in ISO 3166-2 format. Optional.
        latitude:
            Float value representing latitude. Optional.
        longitude:
            Float value representing longitude. Optional.
        results_dict:
            Output dict for geolocation results.
    """

    admin1_result = find_admin1_region(country, admin1_region, latitude, longitude)
    results_dict[filepath] = {
        "country": country,
        "admin1_region": admin1_result,
        "latitude": latitude,
        "longitude": longitude,
    }


def _combine_results(  # pylint: disable=too-many-positional-arguments
    ensemble: SpeciesNetEnsemble,
    filepaths: list[str],
    classifier_results: dict[str, dict],
    detector_results: dict[str, dict],
    geolocation_results: dict[str, dict],
    partial_predictions: dict[str, dict],
    predictions_json: Optional[StrPath] = None,
    save_lock: Optional[threading.Lock] = None,
) -> Optional[dict]:
    """Combines inference results from multiple jobs that ran independently.

    Brings together results from the classifier, detector, and geolocation steps to
    create the final predictions. The SpeciesNet ensemble model is used to combine these
    results, which may be saved to a JSON file.

    Args:
        ensemble:
            SpeciesNetEnsemble to run.
        filepaths:
            List of filepaths to ensemble predictions for.
        classifier_results:
            Dict of classifier results, with keys given by the filepaths to ensemble
            predictions for.
        detector_results:
            Dict of detector results, with keys given by the filepaths to ensemble
            predictions for.
        geolocation_results:
            Dict of geolocation results, with keys given by the filepaths to ensemble
            predictions for.
        partial_predictions:
            Dict of partial predictions from previous ensemblings, with keys given by
            the filepaths for which predictions where already ensembled. Used to skip
            re-ensembling for the matching filepaths.
        predictions_json:
            Output filepath where to save the predictions dict in JSON format. If
            `None`, predictions are not saved to a file and are returned instead.
        save_lock:
            Threading lock used to avoid race conditions when saving predictions to a
            file. Only needed when `predictions_json` is not `None`, otherwise it is
            ignored.

    Returns:
        The predictions dict of ensembled inference results if `predictions_json` is set
        to `None`, otherwise return `None` since predictions are saved to a file.
    """

    ensemble_results = ensemble.combine(
        filepaths=filepaths,
        classifier_results=classifier_results,
        detector_results=detector_results,
        geolocation_results=geolocation_results,
        partial_predictions=partial_predictions,
    )
    predictions_dict = {"predictions": ensemble_results}
    if predictions_json:
        if save_lock:
            with save_lock:
                save_predictions(predictions_dict, predictions_json)
        else:
            save_predictions(predictions_dict, predictions_json)
    else:
        return predictions_dict


def _merge_results(  # pylint: disable=too-many-positional-arguments
    filepaths: list[str],
    new_predictions: dict[str, dict],
    partial_predictions: dict[str, dict],
    failure_type: Failure,
    predictions_json: Optional[StrPath] = None,
    save_lock: Optional[threading.Lock] = None,
) -> Optional[dict]:
    """Merges new inference results with partial predictions from previous runs.

    Args:
        filepaths:
            List of filepaths to merge predictions for.
        new_predictions:
            Dict of new inference results keyed by filepaths.
        partial_predictions:
            Dict of partial predictions from previous runs, keyed by filepaths.
        failure_type:
            Type of failure to report when a prediction is missing.
        predictions_json:
            Output filepath where to save the predictions dict in JSON format. If
            `None`, predictions are not saved to a file and are returned instead.
        save_lock:
            Threading lock used to avoid race conditions when saving predictions to a
            file. Only needed when `predictions_json` is not `None`, otherwise it is
            ignored.

    Returns:
        The predictions dict of merged inference results if `predictions_json` is set
        to `None`, otherwise return `None` since predictions are saved to a file.
    """

    results = []
    for filepath in filepaths:
        # Use the result from previously computed predictions when available.
        if filepath in partial_predictions:
            results.append(partial_predictions[filepath])
            continue

        # Use the new prediction when available, or report a failure.
        if filepath in new_predictions:
            results.append(new_predictions[filepath])
        else:
            results.append(
                {
                    "filepath": filepath,
                    "failures": [failure_type.name],
                }
            )

    predictions_dict = {"predictions": results}
    if predictions_json:
        if save_lock:
            with save_lock:
                save_predictions(predictions_dict, predictions_json)
        else:
            save_predictions(predictions_dict, predictions_json)
    else:
        return predictions_dict


def _start_periodic_results_saving(
    fn: Callable,
    *fn_args,
    **fn_kwargs,
) -> tuple[RepeatedAction, threading.Lock]:
    """Starts periodic results saving every 10 minutes.

    Initiates a background thread to save partial results periodically to avoid losing
    progress during longer inference jobs.

    Args:
        fn:
            Callable for saving results.
        *fn_args:
            Arguments for the callable that saves results.
        **fn_kwargs:
            Keyword arguments for the callable that saves results.

    Returns:
        A tuple made of: (a) a repeated action to save results periodically, and (b) a
        threading lock used to avoid race conditions when saving predictions to a file.
    """

    save_lock = threading.Lock()
    periodic_saver = RepeatedAction(
        interval=600,  # 10 minutes.
        fn=fn,
        *fn_args,
        save_lock=save_lock,
        **fn_kwargs,
    )
    periodic_saver.start()
    return periodic_saver, save_lock


def _stop_periodic_results_saving(periodic_saver: Optional[RepeatedAction]) -> None:
    """Stops periodic results saving.

    Args:
        periodic_saver:
            Repeated action to stop.
    """

    if periodic_saver:
        periodic_saver.stop()
        periodic_saver.join()


def _error_callback(e: Exception) -> None:
    """Error callback to log inference errors.

    Args:
        e: Exception to log.
    """

    logging.error(
        "Unexpected inference error:\n%s",
        "".join(traceback.format_exception(type(e), e, e.__traceback__)),
    )


class SpeciesNet:
    """Main interface for running inference with SpeciesNet.

    Offers a high-level interface to run inference with the SpeciesNet model, supporting
    various input formats. It is designed to handle full predictions (with both detector
    and classifier), classification only, detection only, or ensembling only tasks. It
    can also be run on a single thread, with multiple threads, or with multiple
    processes.
    """

    def __init__(
        self,
        model_name: str,
        *,
        components: Literal["all", "classifier", "detector", "ensemble"] = "all",
        geofence: bool = True,
        target_species_txt: Optional[str] = None,
        combine_predictions_fn: Callable = combine_predictions_for_single_item,
        multiprocessing: bool = False,
    ) -> None:
        """Initializes the SpeciesNet model with specified settings.

        Args:
            model_name:
                String value identifying the model to be loaded.
            components:
                String representing which model components to load and run. One of
                ["all", "classifier", "detector", "ensemble"]. Defaults to "all".
            geofence:
                Whether to enable geofencing during ensemble prediction. Defaults to
                `True`.
            target_species_txt:
                Path to a text file containing the target species to always output
                classification scores for. Optional.
            combine_predictions_fn:
                Function to tell the ensemble how to combine predictions from the
                individual model components (e.g. classifications, detections etc.)
            multiprocessing:
                Whether to enable multiprocessing or not. Defaults to `False`.
        """

        if multiprocessing:
            self.manager = SyncManager()
            self.manager.start()  # pylint: disable=consider-using-with
            if components in ["all", "classifier"]:
                self.classifier = self.manager.Classifier(  # type: ignore
                    model_name, target_species_txt=target_species_txt
                )
            if components in ["all", "detector"]:
                self.detector = self.manager.Detector(model_name)  # type: ignore
            if components in ["all", "ensemble"]:
                self.ensemble = self.manager.Ensemble(  # type: ignore
                    model_name,
                    geofence=geofence,
                    prediction_combiner=combine_predictions_fn,
                )
        else:
            self.manager = None
            if components in ["all", "classifier"]:
                self.classifier = SpeciesNetClassifier(
                    model_name, target_species_txt=target_species_txt
                )
            if components in ["all", "detector"]:
                self.detector = SpeciesNetDetector(model_name)
            if components in ["all", "ensemble"]:
                self.ensemble = SpeciesNetEnsemble(
                    model_name,
                    geofence=geofence,
                    prediction_combiner=combine_predictions_fn,
                )

    def __del__(self) -> None:
        """Cleanup method."""

        if self.manager:
            tqdm.set_lock(threading.RLock())
            if hasattr(self.manager, "shutdown"):
                self.manager.shutdown()

    def _predict_using_single_thread(
        self,
        instances_dict: dict,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        """Runs prediction using a single thread, processing each image one by one.

        This approach is useful for debugging or in environments where parallelization
        is not suitable. All the inference components run sequentially within the same
        thread.

        Args:
            instances_dict:
                 Instances dict to process.
            progress_bars:
                Whether to show progress bars.
            predictions_json:
                Path where to save the JSON output.

        Returns:
             The predictions dict of ensembled inference results if `predictions_json`
             is set to `None`, otherwise return `None` since predictions are saved to
             a file.
        """
        instances = instances_dict["instances"]
        filepaths = [instance["filepath"] for instance in instances]
        classifier_results = {}
        detector_results = {}
        geolocation_results = {}

        # Load previously computed predictions and identify remaining instances to
        # process.
        partial_predictions, instances_to_process = load_partial_predictions(
            predictions_json, instances
        )
        num_instances_to_process = len(instances_to_process)

        # Start a periodic saver if an output file was specified.
        if predictions_json:
            periodic_saver, save_lock = _start_periodic_results_saving(
                _combine_results,
                ensemble=self.ensemble,
                filepaths=filepaths,
                classifier_results=classifier_results,
                detector_results=detector_results,
                geolocation_results=geolocation_results,
                partial_predictions=partial_predictions,
                predictions_json=predictions_json,
            )
        else:
            periodic_saver = None
            save_lock = None

        # Set up progress tracking.
        progress = Progress(
            enabled=(
                [
                    "detector_preprocess",
                    "detector_predict",
                    "classifier_preprocess",
                    "classifier_predict",
                    "geolocation",
                ]
                if progress_bars
                else []
            ),
            total=num_instances_to_process,
            rlock=threading.RLock(),
        )

        # Process instances one by one.
        for instance in instances_to_process:
            filepath = instance["filepath"]
            country = instance.get("country")
            admin1_region = instance.get("admin1_region")
            latitude = instance.get("latitude")
            longitude = instance.get("longitude")

            # Load image.
            img = load_rgb_image(filepath)

            # Preprocess image for detector.
            detector_input = self.detector.preprocess(img)
            progress.update("detector_preprocess")

            # Run detector.
            detector_results[filepath] = self.detector.predict(filepath, detector_input)
            progress.update("detector_predict")

            # Preprocess image for classifier.
            detections = detector_results[filepath].get("detections", None)
            if detections:
                bboxes = [BBox(*det["bbox"]) for det in detections]
            else:
                bboxes = []
            classifier_input = self.classifier.preprocess(img, bboxes=bboxes)
            progress.update("classifier_preprocess")

            # Run classifier.
            classifier_results[filepath] = self.classifier.predict(
                filepath, classifier_input
            )
            progress.update("classifier_predict")

            # Run geolocation.
            admin1_result = find_admin1_region(
                country, admin1_region, latitude, longitude
            )
            geolocation_results[filepath] = {
                "country": country,
                "admin1_region": admin1_result,
                "latitude": latitude,
                "longitude": longitude,
            }
            progress.update("geolocation")

        # Stop progress tracking.
        progress.stop()

        # Stop the periodic saver if an output file was specified.
        if predictions_json:
            _stop_periodic_results_saving(periodic_saver)

        # Ensemble predictions.
        return _combine_results(
            ensemble=self.ensemble,
            filepaths=filepaths,
            classifier_results=classifier_results,
            detector_results=detector_results,
            geolocation_results=geolocation_results,
            partial_predictions=partial_predictions,
            predictions_json=predictions_json,
            save_lock=save_lock,
        )

    def _predict_using_worker_pools(  # pylint: disable=too-many-positional-arguments
        self,
        instances_dict: dict,
        batch_size: int = 8,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
        new_pool_fn: Optional[Callable] = None,
        new_list_fn: Optional[Callable] = None,
        new_dict_fn: Optional[Callable] = None,
        new_queue_fn: Optional[Callable] = None,
        new_rlock_fn: Optional[Callable] = None,
    ) -> Optional[dict]:
        """Runs prediction using worker pools (multi-threading or multiprocessing).

        This method uses worker pools to process images concurrently. This is an
        abstract method that accepts worker pools of different types.

        Args:
            instances_dict:
                Instances dict to process.
            batch_size:
                Batch size for inference.
            progress_bars:
                Whether to show progress bars.
            predictions_json:
                Path where to save the JSON output.
            new_pool_fn:
                Callable that returns a new pool.
            new_list_fn:
                Callable that returns a list.
            new_dict_fn:
                Callable that returns a dict.
            new_queue_fn:
                Callable that returns a queue.
             new_rlock_fn:
                Callable that returns a thread/process lock.

        Returns:
            The predictions dict of ensembled inference results if `predictions_json`
            is set to `None`, otherwise return `None` since predictions are saved to
            a file.
        """
        assert new_pool_fn is not None
        assert new_list_fn is not None
        assert new_dict_fn is not None
        assert new_queue_fn is not None
        assert new_rlock_fn is not None

        instances = instances_dict["instances"]
        filepaths = new_list_fn([instance["filepath"] for instance in instances])
        classifier_results = new_dict_fn()
        detector_results = new_dict_fn()
        geolocation_results = new_dict_fn()

        # Load previously computed predictions and identify remaining instances to
        # process.
        partial_predictions, instances_to_process = load_partial_predictions(
            predictions_json, instances
        )
        partial_predictions = new_dict_fn(partial_predictions)
        num_instances_to_process = len(instances_to_process)
        num_batches = num_instances_to_process // batch_size + min(
            num_instances_to_process % batch_size, 1
        )
        last_batch_size = num_instances_to_process % batch_size
        if not last_batch_size:
            last_batch_size = batch_size

        # Start a periodic saver if an output file was specified.
        if predictions_json:
            periodic_saver, save_lock = _start_periodic_results_saving(
                _combine_results,
                ensemble=self.ensemble,
                filepaths=filepaths,
                classifier_results=classifier_results,
                detector_results=detector_results,
                geolocation_results=geolocation_results,
                partial_predictions=partial_predictions,
                predictions_json=predictions_json,
            )
        else:
            periodic_saver = None
            save_lock = None

        # Set up progress tracking.
        progress = Progress(
            enabled=(
                [
                    "detector_preprocess",
                    "detector_predict",
                    "classifier_preprocess",
                    "classifier_predict",
                    "geolocation",
                ]
                if progress_bars
                else []
            ),
            total=num_instances_to_process,
            batches=num_batches,
            rlock=new_rlock_fn(),
        )

        # Set up multiprocessing pools and queues.
        common_pool = (
            new_pool_fn()
        )  # Limited by the number of logical CPUs on the machine.
        detector_pool = new_pool_fn(1)  # One single worker to run detector inference.
        classifier_pool = new_pool_fn(
            1
        )  # One single worker to run classifier inference.
        detector_queue = new_queue_fn(
            max(2 * batch_size, 64)
        )  # Limited number of images to store in memory.
        bboxes_queue = new_queue_fn()  # Unlimited number of bboxes to store in memory.
        classifier_queue = new_queue_fn(
            max(2 * batch_size, 64)
        )  # Limited number of images to store in memory.

        # Run a bunch of small tasks asynchronously.
        for instance in instances_to_process:

            # Preprocess image for detector.
            common_pool.apply_async(
                _prepare_detector_input,
                args=(self.detector, instance["filepath"], detector_queue),
                callback=lambda _: progress.update("detector_preprocess"),
                error_callback=_error_callback,
            )

            # Preprocess image for classifier.
            common_pool.apply_async(
                _prepare_classifier_input,
                args=(self.classifier, bboxes_queue, classifier_queue),
                callback=lambda _: progress.update("classifier_preprocess"),
                error_callback=_error_callback,
            )

            # Run geolocation.
            common_pool.apply_async(
                _find_admin1_region,
                args=(
                    instance["filepath"],
                    instance.get("country"),
                    instance.get("admin1_region"),
                    instance.get("latitude"),
                    instance.get("longitude"),
                    geolocation_results,
                ),
                callback=lambda _: progress.update("geolocation"),
                error_callback=_error_callback,
            )

        # Run detector inference asynchronously.
        for _ in range(num_instances_to_process):
            detector_pool.apply_async(
                _run_detector,
                args=(self.detector, detector_queue, detector_results, bboxes_queue),
                callback=lambda _: progress.update("detector_predict"),
                error_callback=_error_callback,
            )

        # Run classifier inference asynchronously.
        for batch_idx in range(num_batches):
            classifier_pool.apply_async(
                _run_classifier,
                args=(
                    self.classifier,
                    classifier_queue,
                    classifier_results,
                    (batch_size if batch_idx < num_batches - 1 else last_batch_size),
                ),
                callback=lambda _: progress.update("classifier_predict"),
                error_callback=_error_callback,
            )

        # Wait for all workers to finish.
        common_pool.close()
        detector_pool.close()
        classifier_pool.close()
        common_pool.join()
        detector_pool.join()
        classifier_pool.join()

        # Stop progress tracking.
        progress.stop()

        # Stop the periodic saver if an output file was specified.
        if predictions_json:
            _stop_periodic_results_saving(periodic_saver)

        # Ensemble predictions.
        return _combine_results(
            ensemble=self.ensemble,
            filepaths=filepaths,
            classifier_results=classifier_results,
            detector_results=detector_results,
            geolocation_results=geolocation_results,
            partial_predictions=partial_predictions,
            predictions_json=predictions_json,
            save_lock=save_lock,
        )

    def _predict_using_thread_pools(
        self,
        instances_dict: dict,
        batch_size: int = 8,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        return self._predict_using_worker_pools(
            instances_dict,
            batch_size=batch_size,
            progress_bars=progress_bars,
            predictions_json=predictions_json,
            new_pool_fn=ThreadPool,
            new_list_fn=list,
            new_dict_fn=dict,
            new_queue_fn=queue.Queue,
            new_rlock_fn=threading.RLock,
        )

    def _predict_using_process_pools(
        self,
        instances_dict: dict,
        batch_size: int = 8,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        assert self.manager is not None
        return self._predict_using_worker_pools(
            instances_dict,
            batch_size=batch_size,
            progress_bars=progress_bars,
            predictions_json=predictions_json,
            new_pool_fn=mp.Pool,
            new_list_fn=self.manager.list,
            new_dict_fn=self.manager.dict,
            new_queue_fn=self.manager.Queue,
            new_rlock_fn=self.manager.RLock,
        )

    def _classify_using_worker_pools(  # pylint: disable=too-many-positional-arguments
        self,
        instances_dict: dict,
        detections_dict: Optional[dict] = None,
        batch_size: int = 8,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
        new_pool_fn: Optional[Callable] = None,
        new_dict_fn: Optional[Callable] = None,
        new_queue_fn: Optional[Callable] = None,
        new_rlock_fn: Optional[Callable] = None,
    ) -> Optional[dict]:
        assert new_pool_fn is not None
        assert new_dict_fn is not None
        assert new_queue_fn is not None
        assert new_rlock_fn is not None

        instances = instances_dict["instances"]
        filepaths = [instance["filepath"] for instance in instances]
        detections_dict = detections_dict or {}
        classifier_results = new_dict_fn()

        # Load previously computed predictions and identify remaining instances to
        # process.
        partial_predictions, instances_to_process = load_partial_predictions(
            predictions_json, instances
        )
        partial_predictions = new_dict_fn(partial_predictions)
        num_instances_to_process = len(instances_to_process)
        num_batches = num_instances_to_process // batch_size + min(
            num_instances_to_process % batch_size, 1
        )
        last_batch_size = num_instances_to_process % batch_size
        if not last_batch_size:
            last_batch_size = batch_size

        # Start a periodic saver if an output file was specified.
        if predictions_json:
            periodic_saver, save_lock = _start_periodic_results_saving(
                _merge_results,
                filepaths=filepaths,
                new_predictions=classifier_results,
                partial_predictions=partial_predictions,
                failure_type=Failure.CLASSIFIER,
                predictions_json=predictions_json,
            )
        else:
            periodic_saver = None
            save_lock = None

        # Set up progress tracking.
        progress = Progress(
            enabled=(
                ["classifier_preprocess", "classifier_predict"] if progress_bars else []
            ),
            total=num_instances_to_process,
            batches=num_batches,
            rlock=new_rlock_fn(),
        )

        # Set up multiprocessing pools and queues.
        common_pool = (
            new_pool_fn()
        )  # Limited by the number of logical CPUs on the machine.
        classifier_pool = new_pool_fn(
            1
        )  # One single worker to run classifier inference.
        bboxes_queue = new_queue_fn()  # Unlimited number of bboxes to store in memory.
        classifier_queue = new_queue_fn(
            max(2 * batch_size, 64)
        )  # Limited number of images to store in memory.

        # Preprocess images for classifier.
        for instance in instances_to_process:
            filepath = instance["filepath"]
            detections = detections_dict.get(filepath, {}).get("detections", [])
            bboxes_queue.put((filepath, [BBox(*det["bbox"]) for det in detections]))
            common_pool.apply_async(
                _prepare_classifier_input,
                args=(self.classifier, bboxes_queue, classifier_queue),
                callback=lambda _: progress.update("classifier_preprocess"),
                error_callback=_error_callback,
            )

        # Run classifier.
        for batch_idx in range(num_batches):
            classifier_pool.apply_async(
                _run_classifier,
                args=(
                    self.classifier,
                    classifier_queue,
                    classifier_results,
                    (batch_size if batch_idx < num_batches - 1 else last_batch_size),
                ),
                callback=lambda _: progress.update("classifier_predict"),
                error_callback=_error_callback,
            )

        # Wait for all workers to finish.
        common_pool.close()
        classifier_pool.close()
        common_pool.join()
        classifier_pool.join()

        # Stop progress tracking.
        progress.stop()

        # Stop the periodic saver if an output file was specified.
        if predictions_json:
            _stop_periodic_results_saving(periodic_saver)

        # Return predictions.
        return _merge_results(
            filepaths=filepaths,
            new_predictions=classifier_results,
            partial_predictions=partial_predictions,
            failure_type=Failure.CLASSIFIER,
            predictions_json=predictions_json,
            save_lock=save_lock,
        )

    def _classify_using_thread_pools(  # pylint: disable=too-many-positional-arguments
        self,
        instances_dict: dict,
        detections_dict: Optional[dict] = None,
        batch_size: int = 8,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        return self._classify_using_worker_pools(
            instances_dict,
            detections_dict=detections_dict,
            batch_size=batch_size,
            progress_bars=progress_bars,
            predictions_json=predictions_json,
            new_pool_fn=ThreadPool,
            new_dict_fn=dict,
            new_queue_fn=queue.Queue,
            new_rlock_fn=threading.RLock,
        )

    def _classify_using_process_pools(  # pylint: disable=too-many-positional-arguments
        self,
        instances_dict: dict,
        detections_dict: Optional[dict] = None,
        batch_size: int = 8,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        assert self.manager is not None
        return self._classify_using_worker_pools(
            instances_dict,
            detections_dict=detections_dict,
            batch_size=batch_size,
            progress_bars=progress_bars,
            predictions_json=predictions_json,
            new_pool_fn=mp.Pool,
            new_dict_fn=self.manager.dict,
            new_queue_fn=self.manager.Queue,
            new_rlock_fn=self.manager.RLock,
        )

    def _detect_using_worker_pools(  # pylint: disable=too-many-positional-arguments
        self,
        instances_dict: dict,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
        new_pool_fn: Optional[Callable] = None,
        new_dict_fn: Optional[Callable] = None,
        new_queue_fn: Optional[Callable] = None,
        new_rlock_fn: Optional[Callable] = None,
    ) -> Optional[dict]:
        assert new_pool_fn is not None
        assert new_dict_fn is not None
        assert new_queue_fn is not None
        assert new_rlock_fn is not None

        instances = instances_dict["instances"]
        filepaths = [instance["filepath"] for instance in instances]
        detector_results = new_dict_fn()

        # Load previously computed predictions and identify remaining instances to
        # process.
        partial_predictions, instances_to_process = load_partial_predictions(
            predictions_json, instances
        )
        partial_predictions = new_dict_fn(partial_predictions)
        num_instances_to_process = len(instances_to_process)

        # Start a periodic saver if an output file was specified.
        if predictions_json:
            periodic_saver, save_lock = _start_periodic_results_saving(
                _merge_results,
                filepaths=filepaths,
                new_predictions=detector_results,
                partial_predictions=partial_predictions,
                failure_type=Failure.DETECTOR,
                predictions_json=predictions_json,
            )
        else:
            periodic_saver = None
            save_lock = None

        # Set up progress tracking.
        progress = Progress(
            enabled=(
                ["detector_preprocess", "detector_predict"] if progress_bars else []
            ),
            total=num_instances_to_process,
            rlock=new_rlock_fn(),
        )

        # Set up multiprocessing pools and queues.
        common_pool = (
            new_pool_fn()
        )  # Limited by the number of logical CPUs on the machine.
        detector_pool = new_pool_fn(1)  # One single worker to run detector inference.
        detector_queue = new_queue_fn(
            64
        )  # Limited number of images to store in memory.

        # Preprocess images for detector.
        for instance in instances_to_process:
            common_pool.apply_async(
                _prepare_detector_input,
                args=(self.detector, instance["filepath"], detector_queue),
                callback=lambda _: progress.update("detector_preprocess"),
                error_callback=_error_callback,
            )

        # Run detector.
        for _ in range(num_instances_to_process):
            detector_pool.apply_async(
                _run_detector,
                args=(self.detector, detector_queue, detector_results),
                callback=lambda _: progress.update("detector_predict"),
                error_callback=_error_callback,
            )

        # Wait for all workers to finish.
        common_pool.close()
        detector_pool.close()
        common_pool.join()
        detector_pool.join()

        # Stop progress tracking.
        progress.stop()

        # Stop the periodic saver if an output file was specified.
        if predictions_json:
            _stop_periodic_results_saving(periodic_saver)

        # Return predictions.
        return _merge_results(
            filepaths=filepaths,
            new_predictions=detector_results,
            partial_predictions=partial_predictions,
            failure_type=Failure.DETECTOR,
            predictions_json=predictions_json,
            save_lock=save_lock,
        )

    def _detect_using_thread_pools(
        self,
        instances_dict: dict,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        return self._detect_using_worker_pools(
            instances_dict,
            progress_bars=progress_bars,
            predictions_json=predictions_json,
            new_pool_fn=ThreadPool,
            new_dict_fn=dict,
            new_queue_fn=queue.Queue,
            new_rlock_fn=threading.RLock,
        )

    def _detect_using_process_pools(
        self,
        instances_dict: dict,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        assert self.manager is not None
        return self._detect_using_worker_pools(
            instances_dict,
            progress_bars=progress_bars,
            predictions_json=predictions_json,
            new_pool_fn=mp.Pool,
            new_dict_fn=self.manager.dict,
            new_queue_fn=self.manager.Queue,
            new_rlock_fn=self.manager.RLock,
        )

    def _ensemble_using_single_thread(  # pylint: disable=too-many-positional-arguments
        self,
        instances_dict: dict,
        classifications_dict: Optional[dict] = None,
        detections_dict: Optional[dict] = None,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        instances = instances_dict["instances"]
        filepaths = [instance["filepath"] for instance in instances]
        classifier_results = classifications_dict or {}
        detector_results = detections_dict or {}
        geolocation_results = {}

        # Load previously computed predictions and identify remaining instances to
        # process.
        partial_predictions, instances_to_process = load_partial_predictions(
            predictions_json, instances
        )
        num_instances_to_process = len(instances_to_process)

        # Start a periodic saver if an output file was specified.
        if predictions_json:
            periodic_saver, save_lock = _start_periodic_results_saving(
                _combine_results,
                ensemble=self.ensemble,
                filepaths=filepaths,
                classifier_results=classifier_results,
                detector_results=detector_results,
                geolocation_results=geolocation_results,
                partial_predictions=partial_predictions,
                predictions_json=predictions_json,
            )
        else:
            periodic_saver = None
            save_lock = None

        # Set up progress tracking.
        progress = Progress(
            enabled=(["geolocation"] if progress_bars else []),
            total=num_instances_to_process,
            rlock=threading.RLock(),
        )

        # Process instances one by one.
        for instance in instances_to_process:
            filepath = instance["filepath"]
            country = instance.get("country")
            admin1_region = instance.get("admin1_region")
            latitude = instance.get("latitude")
            longitude = instance.get("longitude")

            # Run geolocation.
            admin1_result = find_admin1_region(
                country, admin1_region, latitude, longitude
            )
            geolocation_results[filepath] = {
                "country": country,
                "admin1_region": admin1_result,
                "latitude": latitude,
                "longitude": longitude,
            }
            progress.update("geolocation")

        # Stop progress tracking.
        progress.stop()

        # Stop the periodic saver if an output file was specified.
        if predictions_json:
            _stop_periodic_results_saving(periodic_saver)

        # Ensemble predictions.
        return _combine_results(
            ensemble=self.ensemble,
            filepaths=filepaths,
            classifier_results=classifier_results,
            detector_results=detector_results,
            geolocation_results=geolocation_results,
            partial_predictions=partial_predictions,
            predictions_json=predictions_json,
            save_lock=save_lock,
        )

    def predict(
        self,
        *,
        instances_dict: Optional[dict] = None,
        instances_json: Optional[StrPath] = None,
        filepaths: Optional[list[StrPath]] = None,
        filepaths_txt: Optional[StrPath] = None,
        folders: Optional[list[StrPath]] = None,
        folders_txt: Optional[StrPath] = None,
        country: Optional[str] = None,
        admin1_region: Optional[str] = None,
        run_mode: Literal[
            "single_thread", "multi_thread", "multi_process"
        ] = "multi_thread",
        batch_size: int = 8,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        instances_dict = prepare_instances_dict(
            instances_dict,
            instances_json,
            filepaths,
            filepaths_txt,
            folders,
            folders_txt,
            country,
            admin1_region,
        )
        if run_mode == "single_thread":
            return self._predict_using_single_thread(
                instances_dict,
                progress_bars=progress_bars,
                predictions_json=predictions_json,
            )
        elif run_mode == "multi_thread":
            return self._predict_using_thread_pools(
                instances_dict,
                batch_size=batch_size,
                progress_bars=progress_bars,
                predictions_json=predictions_json,
            )
        elif run_mode == "multi_process":
            return self._predict_using_process_pools(
                instances_dict,
                batch_size=batch_size,
                progress_bars=progress_bars,
                predictions_json=predictions_json,
            )
        else:
            raise ValueError(f"Unknown run mode: `{run_mode}`")

    def classify(
        self,
        *,
        instances_dict: Optional[dict] = None,
        instances_json: Optional[StrPath] = None,
        filepaths: Optional[list[StrPath]] = None,
        filepaths_txt: Optional[StrPath] = None,
        folders: Optional[list[StrPath]] = None,
        folders_txt: Optional[StrPath] = None,
        country: Optional[str] = None,
        admin1_region: Optional[str] = None,
        detections_dict: Optional[dict] = None,
        run_mode: Literal["multi_thread", "multi_process"] = "multi_thread",
        batch_size: int = 8,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        instances_dict = prepare_instances_dict(
            instances_dict,
            instances_json,
            filepaths,
            filepaths_txt,
            folders,
            folders_txt,
            country,
            admin1_region,
        )
        if run_mode == "multi_thread":
            return self._classify_using_thread_pools(
                instances_dict,
                detections_dict=detections_dict,
                batch_size=batch_size,
                progress_bars=progress_bars,
                predictions_json=predictions_json,
            )
        elif run_mode == "multi_process":
            return self._classify_using_process_pools(
                instances_dict,
                detections_dict=detections_dict,
                batch_size=batch_size,
                progress_bars=progress_bars,
                predictions_json=predictions_json,
            )
        else:
            raise ValueError(f"Unknown run mode: `{run_mode}`")

    def detect(
        self,
        *,
        instances_dict: Optional[dict] = None,
        instances_json: Optional[StrPath] = None,
        filepaths: Optional[list[StrPath]] = None,
        filepaths_txt: Optional[StrPath] = None,
        folders: Optional[list[StrPath]] = None,
        folders_txt: Optional[StrPath] = None,
        country: Optional[str] = None,
        admin1_region: Optional[str] = None,
        run_mode: Literal["multi_thread", "multi_process"] = "multi_thread",
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        instances_dict = prepare_instances_dict(
            instances_dict,
            instances_json,
            filepaths,
            filepaths_txt,
            folders,
            folders_txt,
            country,
            admin1_region,
        )
        if run_mode == "multi_thread":
            return self._detect_using_thread_pools(
                instances_dict,
                progress_bars=progress_bars,
                predictions_json=predictions_json,
            )
        elif run_mode == "multi_process":
            return self._detect_using_process_pools(
                instances_dict,
                progress_bars=progress_bars,
                predictions_json=predictions_json,
            )
        else:
            raise ValueError(f"Unknown run mode: `{run_mode}`")

    def ensemble_from_past_runs(
        self,
        *,
        instances_dict: Optional[dict] = None,
        instances_json: Optional[StrPath] = None,
        filepaths: Optional[list[StrPath]] = None,
        filepaths_txt: Optional[StrPath] = None,
        folders: Optional[list[StrPath]] = None,
        folders_txt: Optional[StrPath] = None,
        country: Optional[str] = None,
        admin1_region: Optional[str] = None,
        classifications_dict: Optional[dict] = None,
        detections_dict: Optional[dict] = None,
        progress_bars: bool = False,
        predictions_json: Optional[StrPath] = None,
    ) -> Optional[dict]:
        instances_dict = prepare_instances_dict(
            instances_dict,
            instances_json,
            filepaths,
            filepaths_txt,
            folders,
            folders_txt,
            country,
            admin1_region,
        )
        return self._ensemble_using_single_thread(
            instances_dict,
            classifications_dict=classifications_dict,
            detections_dict=detections_dict,
            progress_bars=progress_bars,
            predictions_json=predictions_json,
        )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-module-docstring

import numpy as np
from numpy.testing import assert_array_equal
import PIL.Image
import pytest

from speciesnet.classifier import SpeciesNetClassifier
from speciesnet.utils import BBox
from speciesnet.utils import load_rgb_image
from speciesnet.utils import PreprocessedImage

# fmt: off
# pylint: disable=line-too-long

BLANK = "f1856211-cfb7-4a5b-9158-c0f72fd09ee6;;;;;;blank"
HUMAN = "990ae9dd-7a59-4344-afcb-1b7b21368000;mammalia;primates;hominidae;homo;sapiens;human"
VEHICLE = "e2895ed5-780b-48f6-8a11-9e27cb594511;;;;;;vehicle"

AFRICAN_ELEPHANT = "55631055-3e0e-4b7a-9612-dedebe9f78b0;mammalia;proboscidea;elephantidae;loxodonta;africana;african elephant"
AMERICAN_BLACK_BEAR = "436ddfdd-bc43-44c3-a25d-34671d3430a0;mammalia;carnivora;ursidae;ursus;americanus;american black bear"
DOMESTIC_CATTLE = "aca65aaa-8c6d-4b69-94de-842b08b13bd6;mammalia;cetartiodactyla;bovidae;bos;taurus;domestic cattle"
DOMESTIC_DOG = "3d80f1d6-b1df-4966-9ff4-94053c7a902a;mammalia;carnivora;canidae;canis;familiaris;domestic dog"
OCELOT = "22976d14-d424-4f18-a67a-d8e1689cefcc;mammalia;carnivora;felidae;leopardus;pardalis;ocelot"

# pylint: enable=line-too-long
# fmt: on


class TestClassifier:
    """Tests for the classifier component."""

    @pytest.fixture(scope="class")
    def classifier(self, model_name: str) -> SpeciesNetClassifier:
        return SpeciesNetClassifier(model_name)

    @pytest.fixture
    def img_green_w40_h100(self) -> PIL.Image.Image:
        return PIL.Image.new("RGB", (40, 100), color=(0, 255, 0))

    @pytest.fixture
    def img_green_w100_h700(self) -> PIL.Image.Image:
        return PIL.Image.new("RGB", (100, 700), color=(0, 255, 0))

    @pytest.fixture
    def img_red_green_blue_w100_h150_h700_h150(self) -> PIL.Image.Image:
        img_red = PIL.Image.new("RGB", (100, 150), color=(255, 0, 0))
        img_green = PIL.Image.new("RGB", (100, 700), color=(0, 255, 0))
        img_blue = PIL.Image.new("RGB", (100, 150), color=(0, 0, 255))
        img = PIL.Image.new("RGB", (100, 1000))
        img.paste(img_red, (0, 0))
        img.paste(img_green, (0, 150))
        img.paste(img_blue, (0, 850))
        return img

    @pytest.fixture
    def img_green_w100_h2000(self) -> PIL.Image.Image:
        return PIL.Image.new("RGB", (100, 2000), color=(0, 255, 0))

    @pytest.fixture
    def img_red_green_blue_w100_h200_h2000_h200(self) -> PIL.Image.Image:
        img_red = PIL.Image.new("RGB", (100, 200), color=(255, 0, 0))
        img_green = PIL.Image.new("RGB", (100, 2000), color=(0, 255, 0))
        img_blue = PIL.Image.new("RGB", (100, 200), color=(0, 0, 255))
        img = PIL.Image.new("RGB", (100, 2400))
        img.paste(img_red, (0, 0))
        img.paste(img_green, (0, 200))
        img.paste(img_blue, (0, 2200))
        return img

    @pytest.fixture
    def img_green_w480_h480(self) -> PIL.Image.Image:
        return PIL.Image.new("RGB", (480, 480), color=(0, 255, 0))

    def test_preprocess(  # pylint: disable=too-many-positional-arguments
        self,
        classifier,
        img_red_green_blue_w100_h150_h700_h150,  # input
        img_green_w40_h100,  # output
        img_green_w100_h700,  # output
        img_red_green_blue_w100_h200_h2000_h200,  # input
        img_green_w100_h2000,  # output
        img_green_w480_h480,  # output
    ) -> None:

        assert classifier.preprocess(None) is None

        if classifier.model_info.type_ == "always_crop":

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
                resize=False,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (1000, 100, 3)
            assert_array_equal(preprocessed.arr, img_red_green_blue_w100_h150_h700_h150)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
                bboxes=[],
                resize=False,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (1000, 100, 3)
            assert_array_equal(preprocessed.arr, img_red_green_blue_w100_h150_h700_h150)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
                bboxes=[BBox(0 / 100, 150 / 1000, 100 / 100, 700 / 1000)],
                resize=False,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (700, 100, 3)
            assert_array_equal(preprocessed.arr, img_green_w100_h700)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
                bboxes=[BBox(0 / 100, 150 / 1000, 100 / 100, 700 / 1000)],
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (480, 480, 3)
            assert_array_equal(preprocessed.arr, img_green_w480_h480)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
                bboxes=[BBox(10 / 100, 150 / 1000, 40 / 100, 100 / 1000)],
                resize=False,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (100, 40, 3)
            assert_array_equal(preprocessed.arr, img_green_w40_h100)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
                bboxes=[BBox(10 / 100, 150 / 1000, 40 / 100, 100 / 1000)],
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (480, 480, 3)
            assert_array_equal(preprocessed.arr, img_green_w480_h480)

        elif classifier.model_info.type_ == "full_image":

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
                bboxes=[BBox(10 / 100, 150 / 1000, 40 / 100, 100 / 1000)],
                resize=False,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (700, 100, 3)
            assert_array_equal(preprocessed.arr, img_green_w100_h700)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
                resize=False,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (700, 100, 3)
            assert_array_equal(preprocessed.arr, img_green_w100_h700)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h150_h700_h150,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 1000
            assert preprocessed.arr.shape == (480, 480, 3)
            assert_array_equal(preprocessed.arr, img_green_w480_h480)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h200_h2000_h200,
                resize=False,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 2400
            assert preprocessed.arr.shape == (2000, 100, 3)
            assert_array_equal(preprocessed.arr, img_green_w100_h2000)

            preprocessed = classifier.preprocess(
                img_red_green_blue_w100_h200_h2000_h200,
            )
            assert preprocessed
            assert preprocessed.orig_width == 100
            assert preprocessed.orig_height == 2400
            assert preprocessed.arr.shape == (480, 480, 3)
            assert_array_equal(preprocessed.arr, img_green_w480_h480)

    def test_predict(self, classifier, img_green_w480_h480) -> None:

        filepath = "missing.jpg"
        prediction = classifier.predict(filepath, None)
        assert prediction["filepath"] == filepath
        assert "failures" in prediction

        filepath = "green.png"
        prediction = classifier.predict(
            filepath, PreprocessedImage(np.asarray(img_green_w480_h480), 480, 480)
        )
        assert prediction["filepath"] == filepath
        assert "failures" not in prediction
        assert "classifications" in prediction
        assert "classes" in prediction["classifications"]
        assert "scores" in prediction["classifications"]
        assert prediction["classifications"]["scores"] == sorted(
            prediction["classifications"]["scores"], reverse=True
        )

    @pytest.fixture(
        # Expected classifications.
        params=[
            (
                "test_data/blank.jpg",
                [],
                BLANK,
            ),
            (
                "test_data/blank2.jpg",
                [],
                BLANK,
            ),
            (
                "test_data/blank3.jpg",
                [BBox(0.0009191, 0, 0.996, 0.9571)],
                BLANK,
            ),
            (
                "test_data/african_elephants.jpg",
                [BBox(0.7041, 0.4765, 0.1108, 0.125)],
                AFRICAN_ELEPHANT,
            ),
            (
                "test_data/african_elephants_bw.jpg",
                [BBox(0.7031, 0.4765, 0.1118, 0.1256)],
                AFRICAN_ELEPHANT,
            ),
            (
                "test_data/african_elephants_cmyk.jpg",
                [BBox(0.7041, 0.4765, 0.1108, 0.125)],
                AFRICAN_ELEPHANT,
            ),
            (
                "test_data/african_elephants_truncated_file.jpg",
                [BBox(0.7045, 0.4759, 0.1103, 0.1256)],
                AFRICAN_ELEPHANT,
            ),
            (
                "test_data/african_elephants_with_exif_orientation.jpg",
                [BBox(0.7041, 0.4765, 0.1108, 0.125)],
                AFRICAN_ELEPHANT,
            ),
            (
                "test_data/american_black_bear.jpg",
                [BBox(0.4331, 0.4283, 0.3232, 0.3222)],
                AMERICAN_BLACK_BEAR,
            ),
            (
                "test_data/domestic_cattle.jpg",
                [BBox(0.3803, 0.4348, 0.1176, 0.1575)],
                DOMESTIC_CATTLE,
            ),
            (
                "test_data/domestic_dog.jpg",
                [BBox(0.2377, 0.08398, 0.5161, 0.6497)],
                DOMESTIC_DOG,
            ),
            (
                "test_data/human.jpg",
                [BBox(0.7115, 0.4976, 0.0664, 0.2424)],
                HUMAN,
            ),
            (
                "test_data/ocelot.jpg",
                [BBox(0.0478, 0.5548, 0.521, 0.2743)],
                OCELOT,
            ),
            (
                "test_data/vehicle.jpg",
                [BBox(0, 0.02083, 1.0, 0.6816)],
                VEHICLE,
            ),
        ]
    )
    def predicted_vs_expected(self, classifier, request) -> tuple[dict, str]:
        filepath, bboxes, label = request.param
        img = classifier.preprocess(load_rgb_image(filepath), bboxes=bboxes)
        assert img is not None
        return classifier.predict(filepath, img)["classifications"], label

    def test_classifications(self, predicted_vs_expected) -> None:
        classifications, label = predicted_vs_expected
        assert classifications["classes"][0] == label
        assert classifications["scores"] == sorted(
            classifications["scores"], reverse=True
        )
